"""Tests for the actors.packagecloud package"""

import datetime
import mock

from tornado import testing

from kingpin.actors import exceptions
from kingpin.actors import packagecloud
from kingpin.actors.test.helper import mock_tornado, tornado_value

__author__ = 'Charles McLaughlin <charles@nextdoor.com>'

ALL_PACKAGES_MOCK_RESPONSE = [
    {'name': 'unittest', 'uploader_name': 'unittest',
     'created_at': '2015-07-07T20:27:18.000Z',
     'distro_version': 'ubuntu/trusty',
     'filename': 'unittest_0.2-1_all.deb', 'epoch': 0,
     'version': '0.2', 'private': True, 'release': '1',
     'package_url': ('/api/v1/repos/unittest/test/package/deb/ubuntu/'
                     'trusty/unittest/all/0.2-1.json'),
     'type': 'deb',
     'package_html_url': ('/unittest/test/packages/ubuntu/trusty/'
                          'unittest_0.2-1_all.deb'),
     'repository_html_url': '/unittest/test'},

    {'name': 'unittest', 'uploader_name': 'unittest',
     'created_at': '2014-07-07T20:27:18.000Z',
     'distro_version': 'ubuntu/trusty',
     'filename': 'unittest_0.1-1_all.deb', 'epoch': 0,
     'version': '0.1', 'private': True, 'release': '1',
     'package_url': ('/api/v1/repos/unittest/test/package/deb/ubuntu/'
                     'trusty/unittest/all/0.1-1.json'),
     'type': 'deb',
     'package_html_url': ('/unittest/test/packages/ubuntu/trusty/'
                          'unittest_0.1-1_all.deb'),
     'repository_html_url': '/unittest/test'}
]


def _get_older_than():
    """ Method for getting an `older_than` value which is used in a few
    tests below.

    The tests specifically look for the second package in
    ALL_PACKAGES_MOCK_RESPONSE, as it's older than the first"""

    created_at = datetime.datetime.strptime(
        ALL_PACKAGES_MOCK_RESPONSE[0]['created_at'],
        '%Y-%m-%dT%H:%M:%S.%fZ')
    older_than = datetime.datetime.now() - created_at + datetime.timedelta(
        seconds=600)  # Adding 5m in case the tests run long
    return older_than


class TestPackagecloudBase(testing.AsyncTestCase):

    """Unit tests for the packagecloud Base actor."""

    def setUp(self, *args, **kwargs):
        super(TestPackagecloudBase, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = 'Unittest'
        packagecloud.ACCOUNT = 'Unittest'
        self.maxDiff = None

    @testing.gen_test
    def test_init_missing_token(self):
        # Un-set the token and make sure the init fails
        packagecloud.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            packagecloud.PackagecloudBase('Unit Test Action', {})

    @testing.gen_test
    def test_init_missing_account(self):
        # Un-set the account and make sure the init fails
        packagecloud.ACCOUNT = None
        with self.assertRaises(exceptions.InvalidCredentials):
            packagecloud.PackagecloudBase('Unit Test Action', {})

    @testing.gen_test
    def test_get_all_packages(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        all_packages = yield actor._get_all_packages(repo='unittest')
        self.assertEqual(all_packages, ALL_PACKAGES_MOCK_RESPONSE)

    @testing.gen_test
    def test_get_package_versions(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        packages = yield actor._get_all_packages(repo='unittest')
        versions = actor._get_package_versions(
            name='unittest', packages=packages)

        self.assertEqual(
            versions,
            [{'created_at': datetime.datetime(2014, 7, 7, 20, 27, 18),
              'name': 'unittest',
              'distro_version': 'ubuntu/trusty',
              'filename': 'unittest_0.1-1_all.deb'},
             {'created_at': datetime.datetime(2015, 7, 7, 20, 27, 18),
              'name': 'unittest',
              'distro_version': 'ubuntu/trusty',
              'filename': 'unittest_0.2-1_all.deb'}])

    @testing.gen_test
    def test_filter_packages(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        packages = yield actor._get_all_packages(repo='unittest')
        packages_list_to_delete = actor._filter_packages(
            regex='unittest', packages=packages)
        self.assertEqual(packages_list_to_delete, set(['unittest']))

    @testing.gen_test
    def test_delete(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._delete(
            regex='unittest', repo='unittest')

        self.assertEqual(deleted_packages,
                         [{'created_at': datetime.datetime(
                             2014, 7, 7, 20, 27, 18),
                             'name': 'unittest',
                             'distro_version': 'ubuntu/trusty',
                             'filename': 'unittest_0.1-1_all.deb'},
                             {'created_at': datetime.datetime(
                                 2015, 7, 7, 20, 27, 18),
                                 'name': 'unittest',
                                 'distro_version': 'ubuntu/trusty',
                                 'filename': 'unittest_0.2-1_all.deb'}])

    @testing.gen_test
    def test_delete_dry(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})
        actor._dry = True

        deleted_packages = yield actor._delete(
            regex='unittest', repo='unittest')

        self.assertEqual(deleted_packages,
                         [{'created_at': datetime.datetime(
                             2014, 7, 7, 20, 27, 18),
                             'name': 'unittest',
                             'distro_version': 'ubuntu/trusty',
                             'filename': 'unittest_0.1-1_all.deb'},
                             {'created_at': datetime.datetime(
                                 2015, 7, 7, 20, 27, 18),
                                 'name': 'unittest',
                                 'distro_version': 'ubuntu/trusty',
                                 'filename': 'unittest_0.2-1_all.deb'}])

    @testing.gen_test
    def test_delete_keep_one(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._delete(
            regex='unittest', repo='unittest', number_to_keep=1)

        self.assertEqual(
            deleted_packages,
            [{'created_at': datetime.datetime(2014, 7, 7, 20, 27, 18),
              'name': 'unittest', 'distro_version': 'ubuntu/trusty',
              'filename': 'unittest_0.1-1_all.deb'}])

    @testing.gen_test
    def test_delete_older_than(self):
        actor = packagecloud.PackagecloudBase('Unit test action', {})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        older_than = _get_older_than()

        deleted_packages = yield actor._delete(
            regex='unittest', repo='unittest',
            older_than=older_than.total_seconds())

        self.assertEqual(
            deleted_packages,
            [{'created_at': datetime.datetime(2014, 7, 7, 20, 27, 18),
              'name': 'unittest', 'distro_version': 'ubuntu/trusty',
              'filename': 'unittest_0.1-1_all.deb'}])


class TestDelete(testing.AsyncTestCase):

    """Unit tests for the packagecloud Delete actor."""

    def setUp(self, *args, **kwargs):
        super(TestDelete, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = 'Unittest'
        packagecloud.ACCOUNT = 'Unittest'
        self.maxDiff = None

    @testing.gen_test
    def test_bad_regex_packages_to_delete(self):
        with self.assertRaises(exceptions.InvalidOptions):
            packagecloud.Delete(
                'Unit test action',
                {'packages_to_delete': '[', 'repo': 'unittest'})

    @testing.gen_test
    def test_execute(self):
        actor = packagecloud.Delete(
            'Unit test action',
            {'packages_to_delete': 'unittest', 'repo': 'unittest'})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._execute()

        self.assertEqual(deleted_packages, None)


class TestDeleteByDate(testing.AsyncTestCase):

    """Unit tests for the packagecloud DeleteByDate actor."""

    def setUp(self, *args, **kwargs):
        super(TestDeleteByDate, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = 'Unittest'
        packagecloud.ACCOUNT = 'Unittest'
        self.maxDiff = None

    @testing.gen_test
    def test_execute(self):
        older_than = _get_older_than()
        actor = packagecloud.DeleteByDate(
            'Unit test action',
            {'packages_to_delete': 'unittest',
             'repo': 'unittest',
             'older_than': int(older_than.total_seconds())})

        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        deleted_packages = yield actor._execute()

        self.assertEqual(deleted_packages, None)


class TestWaitForPackage(testing.AsyncTestCase):

    """Unit tests for the packagecloud WaitForPackage actor."""

    def setUp(self, *args, **kwargs):
        super(TestWaitForPackage, self).setUp(*args, **kwargs)
        packagecloud.TOKEN = 'Unittest'
        packagecloud.ACCOUNT = 'Unittest'
        self.maxDiff = None

    @testing.gen_test
    def test_bad_regex_name(self):
        with self.assertRaises(exceptions.InvalidOptions):
            packagecloud.WaitForPackage(
                'Unit test action',
                {'name': '[', 'version': '1', 'repo': 'unittest'})

    @testing.gen_test
    def test_bad_regex_version(self):
        with self.assertRaises(exceptions.InvalidOptions):
            packagecloud.WaitForPackage(
                'Unit test action',
                {'name': 'unittest', 'version': '[', 'repo': 'unittest'})

    @testing.gen_test
    def test_execute(self):
        actor = packagecloud.WaitForPackage(
            'Unit test action',
            {'name': 'unittest', 'repo': 'unittest', 'version': '0.2'})

        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._packagecloud_client.delete().http_delete = mock_tornado({})

        matched_packages = yield actor._execute()

        self.assertEqual(matched_packages, None)

    @testing.gen_test
    def test_execute_with_sleep(self):
        actor = packagecloud.WaitForPackage(
            'Unit test action',
            {'name': 'not_found', 'repo': 'unittest', 'version': '0.2',
             'sleep': 1})

        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)
        actor._search = mock.Mock(
            side_effect=[tornado_value([]), tornado_value(['something'])])
        yield actor._execute()
        self.assertEqual(actor._search.call_count, 2)

    @testing.gen_test
    def test_search(self):
        actor = packagecloud.WaitForPackage('Unit test action', {
            'name': 'unittest', 'repo': 'unittest', 'version': '0.2'})
        actor._packagecloud_client = mock.Mock()
        actor._packagecloud_client.packages().http_get = mock_tornado(
            ALL_PACKAGES_MOCK_RESPONSE)

        matched_packages = yield actor._search(
            repo='unittest', name='unittest', version='0.2')

        self.assertEqual(matched_packages, [ALL_PACKAGES_MOCK_RESPONSE[0]])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.server_array`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. _ResourceInstances:
   http://reference.rightscale.com/api1.5/resources/
   ResourceInstances.html#update
"""

import logging
import math

from tornado import gen
import mock
import requests

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.rightscale import api
from kingpin.actors.rightscale import base
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


class InvalidInputs(exceptions.InvalidOptions):

    """Raised when supplied inputs are invalid for a ServerArray."""


class TaskExecutionFailed(exceptions.RecoverableActorFailure):

    """Raised when one or more RightScale Task executions fail."""


class ServerArrayBaseActor(base.RightScaleBaseActor):

    """Abstract ServerArray Actor that provides some utility methods."""

    @gen.coroutine
    def _apply(self, function, arrays, *args, **kwargs):
        """Yield a function on several arrays at once.

        Many of our rightscale.server_array Actors have the ability to act on
        multiple arrays at a time (through the 'exact=False' parameter). This
        method provides a quick and re-usable method for yielding generators on
        an array (or group of arrays). All we do here is queue up a group of
        functions, yield them all at once, and return.

        args:
            function: Reference to the function to execute
            arrays: An array, or list of arrays to execute on.
            *args: Any *args to pass to the function
            **kwargs: Any **kwargs to pass to the function
        """
        if not isinstance(arrays, list):
            arrays = [arrays]

        tasks = []
        for array in arrays:
            self.log.debug('Adding %s(%s, %s, %s) to async call list' %
                           (function.__name__, array.soul['name'],
                            args, kwargs))
            tasks.append(function(array, *args, **kwargs))

        self.log.debug('Calling all functions in async call list')
        ret = yield tasks
        raise gen.Return(ret)


class Clone(ServerArrayBaseActor):

    """Clones a RightScale Server Array.

    Clones a ServerArray in RightScale and renames it to the newly supplied
    name.  By default, this actor is extremely strict about validating that the
    ``source`` array already exists, and that the ``dest`` array does not yet
    exist. This behavior can be overridden though if your Kingpin script
    creates the ``source``, or destroys an existing ``dest`` ServerArray
    sometime before this actor executes.

    **Options**

    :source:
      The name of the ServerArray to clone

    :strict_source:
      Whether or not to fail if the source ServerArray does not exist.
      (default: True)

    :dest:
      The new name for your cloned ServerArray

    :strict_dest:
      Whether or not to fail if the destination ServerArray already exists.
      (default: True)

    **Examples**

    Clone my-template-array to my-new-array:

    .. code-block:: json

       { "desc": "Clone my array",
         "actor": "rightscale.server_array.Clone",
         "options": {
           "source": "my-template-array",
           "dest": "my-new-array"
         }
       }

    Clone an array that was created sometime earlier in the Kingpin JSON,
    and thus does not exist yet during the dry run:

    .. code-block:: json

       { "desc": "Clone that array we created earlier",
         "actor": "rightscale.server_array.Clone",
         "options": {
           "source": "my-template-array",
           "strict_source": false,
           "dest": "my-new-array"
         }
       }

    Clone an array into a destination name that was destroyed sometime
    earlier in the Kingpin JSON:

    .. code-block:: json

       { "desc": "Clone that array we created earlier",
         "actor": "rightscale.server_array.Clone",
         "options": {
           "source": "my-template-array",
           "dest": "my-new-array",
           "strict_dest": false,
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* validate that the ``source`` array exists. If
    it does not, a `kingpin.actors.rightscale.api.ServerArrayException` is
    thrown. Once that has been validated, the dry mode execution pretends to
    copy the array by creating a mocked cloned array resource. This mocked
    resource is then operated on during the rest of the execution of the actor,
    guaranteeing that no live resources are modified.

    Example *dry* output::

        [Copy Test (DRY Mode)] Verifying that array "temp" exists
        [Copy Test (DRY Mode)] Verifying that array "new" does not exist
        [Copy Test (DRY Mode)] Cloning array "temp"
        [Copy Test (DRY Mode)] Renaming array "<mocked clone of temp>" to "new"
    """

    all_options = {
        'source': (str, REQUIRED, 'Name of the ServerArray to clone.'),
        'strict_source': (bool, True, 'Strict Source ServerArray validation.'),
        'strict_dest': (bool, True, 'Strict Dest ServerArray validation.'),
        'dest': (str, REQUIRED, 'Name to give the cloned ServerArray.')
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Clone, self).__init__(*args, **kwargs)
        # By default, we're strict on our source/dest array validation
        self._source_raise_on = 'notfound'
        self._source_allow_mock = False
        self._dest_raise_on = 'found'
        self._dest_allow_mock = False

        if not self.option('strict_source'):
            self._source_raise_on = None
            self._source_allow_mock = True

        if not self.option('strict_dest'):
            self._dest_raise_on = None
            self._dest_allow_mock = True

    @gen.coroutine
    def _execute(self):
        # Find the array we're copying from
        source_array = yield self._find_server_arrays(
            self.option('source'),
            raise_on=self._source_raise_on,
            allow_mock=self._source_allow_mock)

        # Sanity-check -- make sure that the destination server array doesn't
        # already exist. If it does, bail out!
        yield self._find_server_arrays(
            self.option('dest'),
            raise_on=self._dest_raise_on,
            allow_mock=self._dest_allow_mock)

        # Now, clone the array!
        self.log.info('Cloning array "%s"' % source_array.soul['name'])
        if not self._dry:
            # We're really doin this!
            new_array = yield self._client.clone_server_array(source_array)
        else:
            # In dry run mode. Don't really clone the array, instead we create
            # a mock object and pass that back as if its the new array.
            new_array = mock.MagicMock(name=self.option('dest'))
            new_array_name = '<mocked clone of %s>' % self.option('source')
            new_array.soul = {'name': new_array_name}

        # Lastly, rename the array
        params = self._generate_rightscale_params(
            'server_array', {'name': self.option('dest')})
        self.log.info('Renaming array "%s" to "%s"' % (new_array.soul['name'],
                                                       self.option('dest')))
        yield self._client.update(new_array, params)
        raise gen.Return()


class Update(ServerArrayBaseActor):

    """Update ServerArray Settings

    Updates an existing ServerArray in RightScale with the supplied parameters.
    Can update any parameter that is described in the RightScale API docs here:

    Parameters are passed into the actor in the form of a dictionary, and are
    then converted into the RightScale format. See below for examples.

    **Options**

    :array:
      (str) The name of the ServerArray to update

    :exact:
      (bool) whether or not to search for the exact array name.
      (default: `true`)

    :params:
      (dict) Dictionary of parameters to update

    :inputs:
      (dict) Dictionary of next-instance server arryay inputs to update

    **Examples**

    .. code-block:: json

       { "desc": "Update my array",
         "actor": "rightscale.server_array.Update",
         "options": {
           "array": "my-new-array",
           "params": {
             "elasticity_params": {
               "bounds": {
                 "min_count": 4
               },
               "schedule": [
                 {"day": "Sunday", "max_count": 2,
                  "min_count": 1, "time": "07:00" },
                 {"day": "Sunday", "max_count": 2,
                  "min_count": 2, "time": "09:00" }
               ]
             },
             "name": "my-really-new-name"
           }
         }
       }

    .. code-block:: json

       { "desc": "Update my array inputs",
         "actor": "rightscale.server_array.Update",
         "options": {
           "array": "my-new-array",
           "inputs": {
             "ELB_NAME": "text:foobar"
           }
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* search for the ``array``, but allows it to be
    missing because its highly likely that the array does not exist yet. If the
    array does not exist, a mocked array object is created for the rest of the
    execution.

    During the rest of the execution, the code bypasses making any real changes
    and just tells you what changes it would have made.

    *This means that the dry mode cannot validate that the supplied inputs will
    work.*

    Example *dry* output::

       [Update Test (DRY Mode)] Verifying that array "new" exists
       [Update Test (DRY Mode)] Array "new" not found -- creating a mock.
       [Update Test (DRY Mode)] Would have updated "<mocked array new>" with
       params: {'server_array[name]': 'my-really-new-name',
                'server_array[elasticity_params][bounds][min_count]': '4'}
    """

    all_options = {
        'array': (str, REQUIRED, 'ServerArray name to Update'),
        'exact': (bool, True, (
            'Whether to search for multiple ServerArrays and act on them.')),
        'params': (dict, {}, 'ServerArray RightScale parameters'),
        'inputs': (dict, {}, 'ServerArray inputs for launching.')
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Update, self).__init__(*args, **kwargs)
        self._params = self._generate_rightscale_params(
            'server_array', self.option('params'))
        self._inputs = self._generate_rightscale_params(
            'inputs', self.option('inputs'))

    @gen.coroutine
    def _check_array_inputs(self, array, inputs):
        """Checks the inputs supplied against the ServerArray being updated.

        Verifies that the supplied inputs are actually found in the ServerArray
        that we are going to be updating.

        Raises:
            InvalidInputs()
        """
        # Quick sanity check, make sure we weren't handed a mock object created
        # by the _find_server_array() method. If we were, then the inputs are
        # not checkable. Just warn, and move on.
        if 'fake' in array.soul:
            self.log.warning('Cannot check inputs for non-existent array.')
            raise gen.Return()

        all_inputs = yield self._client.get_server_array_inputs(array)
        all_input_names = [i.soul['name'] for i in all_inputs]

        success = True
        for input_name, _ in list(inputs.items()):
            # Inputs have to be there. If not -- it's a problem.
            if input_name not in all_input_names:
                self.log.error('Input not found: "%s"' % input_name)
                success = False

        if not success:
            raise InvalidInputs('Some inputs supplied were incorrect.')

        raise gen.Return()

    @gen.coroutine
    def _update_params(self, array):
        """Update the parameters on a RightScale ServerArray.

        args:
            array: The array to operate on
        """

        if not self.option('params'):
            raise gen.Return()

        self.log.info('Updating array "%s" with params: %s' %
                      (array.soul['name'], self._params))
        try:
            yield self._client.update(array, self._params)
        except api.RightScaleError as e:
            raise exceptions.RecoverableActorFailure(e)
        except requests.exceptions.HTTPError as e:
            raise exceptions.UnrecoverableActorFailure(e)

        raise gen.Return()

    @gen.coroutine
    def _update_inputs(self, array):
        """Update the inputs on a RightScale ServerArray.

        args:
            array: rightscale.Resource ServerArray Object
        """

        if not self.option('inputs'):
            raise gen.Return()

        self.log.info('Updating array "%s" with inputs: %s' %
                      (array.soul['name'], self._inputs))
        yield self._client.update_server_array_inputs(array, self._inputs)

    @gen.coroutine
    def _execute(self):
        # First, find the arrays we're going to be patching.
        arrays = yield self._find_server_arrays(
            self.option('array'), exact=self.option('exact'))

        # In dry run, just comment that we would have made the change.
        if self._dry:
            self.log.debug('Not making any changes.')
            if self.option('params'):
                self.log.info('Params would be: %s' % self.option('params'))
            if self.option('inputs'):
                self.log.info('Inputs would be: %s' % self.option('inputs'))
                yield self._apply(self._check_array_inputs,
                                  arrays, self.option('inputs'))

            raise gen.Return()

        # Do the real work
        yield self._apply(self._update_params, arrays)
        yield self._apply(self._update_inputs, arrays)
        raise gen.Return()


class UpdateNextInstance(ServerArrayBaseActor):

    """Update the Next Instance parameters for a Server Array

    Updates an existing ServerArray in RightScale with the supplied parameters.
    Can update any parameter that is described in the RightScale
    `ResourceInstances`_ docs.

    **Note about the image_href parameter**

    If you pass in the string `default` to the `image_href` key in your
    `params` dictionary, we will search and find the default image that your
    ServerArray's Multi Cloud Image refers to. This helper is useful if you
    update your ServerArrays to use custom AMIs, and then occasionally want to
    go back to using a stock AMI. For example, if you boot up your instances
    occasionally off a stock AMI, customize the host, and then bake that host
    into a custom AMI.

    Parameters are passed into the actor in the form of a dictionary, and are
    then converted into the RightScale format. See below for examples.

    **Options**

    :array:
      (str) The name of the ServerArray to update

    :exact:
      (bool) whether or not to search for the exact array name.
      (default: `true`)

    :params:
      (dict) Dictionary of parameters to update

    **Examples**

    .. code-block:: json

       { "desc": "Update my array",
         "actor": "rightscale.server_array.UpdateNextInstance",
         "options": {
           "array": "my-new-array",
           "params": {
             "associate_public_ip_address": true,
             "image_href": "/image/href/123",
           }
         }
       }

    .. code-block:: json

       { "desc": "Reset the AMI image to the MCI default",
         "actor": "rightscale.server_array.UpdateNextInstance",
         "options": {
           "array": "my-new-array",
           "params": {
             "image_href": "default",
           }
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* search for the ``array``, but allows it to be
    missing because its highly likely that the array does not exist yet. If the
    array does not exist, a mocked array object is created for the rest of the
    execution.

    During the rest of the execution, the code bypasses making any real changes
    and just tells you what changes it would have made.

    *This means that the dry mode cannot validate that the supplied params will
    work.*

    Example *dry* output::

       [Update my array (DRY Mode)] Verifying that array "new" exists
       [Update my array (DRY Mode)] Array "new" not found -- creating a mock.
       [Update my array (DRY Mode)] Would have updated "<mocked array new>"
       with params: {'server_array[associate_public_ip_address]': true,
                'server_array[image_href]': '/image/href/'}
    """

    all_options = {
        'array': (str, REQUIRED, 'ServerArray name to Update'),
        'exact': (bool, True, (
            'Whether to search for multiple ServerArrays and act on them.')),
        'params': (dict, REQUIRED, 'Next Instance RightScale parameters'),
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(UpdateNextInstance, self).__init__(*args, **kwargs)
        self._params = self._generate_rightscale_params(
            'instance', self.option('params'))

    @gen.coroutine
    def _update_params(self, array):
        """Update the parameters on a RightScale Instance.

        args:
            array: The ServerArray to operate on
        """
        # Get the 'next instance' of the array that we're going to work on
        instance = yield self._client.show(array.next_instance)

        # Get our parameters
        params = self.option('params')

        # Magic: If a user supplies 'default' to the image_href then we do some
        # digging for them and find the 'default' AMI HREF for that server
        # array.
        if ('image_href' in params and params['image_href'] == 'default'):
            params['image_href'] = yield self._find_def_image_href(instance)

        # Second pass at the generating the parameters. We did this at
        # instantiation time as a sanity check to make sure the parameters were
        # half-decent. Now we run it a second time in case any the 'image_href'
        # magic above was executed.
        rs_params = self._generate_rightscale_params(
            'instance', self.option('params'))

        if self._dry:
            self.log.info('Would have updated array\'s next_instance "%s" '
                          'with params: %s' %
                          (instance.soul['name'], rs_params))
            raise gen.Return()

        self.log.info('Updating array\'s next_instance "%s" with params: %s' %
                      (instance.soul['name'], rs_params))

        try:
            yield self._client.update(instance, rs_params)
        except api.RightScaleError as e:
            raise exceptions.RecoverableActorFailure(e)
        except requests.exceptions.HTTPError as e:
            raise exceptions.UnrecoverableActorFailure(e)

        raise gen.Return()

    @gen.coroutine
    def _find_def_image_href(self, instance):
        self.log.debug('Searching for default boot AMI for %s' %
                       instance.soul['name'])

        # Find the MultiCloudImage associated with this 'instance' object, then
        # get the full list of 'settings' for that MCI.
        mci = yield self._client.show(instance.multi_cloud_image)
        self.log.debug('Got MCI: %s' % mci.soul['name'])
        mci_settings = yield self._client.show(mci.settings)
        self.log.debug('Got %s MCI Cloud Settings.' % len(mci_settings))

        # Now, find the 'setting' that matches the cloud of our instance. Note,
        # there should never be more than one returned -- so we take the first
        # one in the list and save it.
        try:
            setting = [s for s in mci_settings if
                       s.cloud.path == instance.cloud.path][0]
            image_href = [l['href'] for l in setting.soul['links']
                          if l['rel'] == 'image'][0]
        except KeyError:
            raise InvalidInputs(
                'Unable to locate default image_href for %s.' % instance.soul)

        raise gen.Return(image_href)

    @gen.coroutine
    def _execute(self):
        # First, find the arrays we're going to be patching.
        arrays = yield self._find_server_arrays(
            self.option('array'), exact=self.option('exact'))

        yield self._apply(self._update_params, arrays)


class Terminate(ServerArrayBaseActor):

    """Terminate all instances in a ServerArray

    Terminates all instances for a ServerArray in RightScale marking the array
    disabled.

    **Options**

    :array:
      (str) The name of the ServerArray to destroy

    :exact:
      (bool) Whether or not to search for the exact array name.
      (default: `true`)

    :strict:
      (bool) Whether or not to fail if the ServerArray does not exist.
      (default: `true`)

    **Examples**

    .. code-block:: json

        { "desc": "Terminate my array",
         "actor": "rightscale.server_array.Terminate",
         "options": {
           "array": "my-array"
         }
       }

    .. code-block:: json

       { "desc": "Terminate many arrays",
         "actor": "rightscale.server_array.Terminate",
         "options": {
           "array": "array-prefix",
           "exact": false,
         }
       }

    **Dry Mode**

    Dry mode still validates that the server array you want to terminate is
    actually gone. If you want to bypass this check, then set the
    ``warn_on_failure`` flag for the actor.
    """

    all_options = {
        'array': (str, REQUIRED, 'ServerArray name to Terminate'),
        'exact': (bool, True, (
            'Whether to search for multiple ServerArrays and act on them.')),
        'strict': (bool, True, 'Strict ServerArray validation.'),
    }

    def __init__(self, *args, **kwargs):
        """Validate the user-supplied parameters at instantiation time."""
        super(Terminate, self).__init__(*args, **kwargs)
        # By default, we're strict on our source/dest array validation
        self._raise_on = 'notfound'
        self._allow_mock = False

        if not self.option('strict'):
            self._raise_on = None
            self._allow_mock = True

    @gen.coroutine
    def _terminate_all_instances(self, array):
        if self._dry:
            self.log.info('Would have terminated all array "%s" instances.' %
                          array.soul['name'])
            raise gen.Return()

        self.log.info('Terminating all instances in array "%s"' %
                      array.soul['name'])
        task = yield self._client.terminate_server_array_instances(array)
        # We don't care if it succeeded -- the multi-terminate job
        # fails all the time when there are hosts still in a
        # 'terminated state' when this call is made. Just wait for it to
        # finish.
        yield self._client.wait_for_task(task)

        raise gen.Return()

    @gen.coroutine
    def _wait_until_empty(self, array, sleep=60):
        """Sleep until all array instances are terminated.

        This loop monitors the server array for its current live instance count
        and waits until the count hits zero before progressing.

        TODO: Add a timeout setting.

        Args:
            array: rightscale.Resource array object
            sleep: Integer time to sleep between checks (def: 60)
        """
        if self._dry:
                self.log.info('Pretending that array %s instances '
                              'are terminated.' % array.soul['name'])
                raise gen.Return()

        while True:
            instances = yield self._client.get_server_array_current_instances(
                array)
            count = len(instances)
            self.log.info('%s instances found' % count)

            if count < 1:
                raise gen.Return()

            # At this point, sleep
            self.log.debug('Sleeping..')
            yield utils.tornado_sleep(sleep)

    @gen.coroutine
    def _disable_array(self, array):
        """Prevent the supplied ServerArray from auto scaling.

        args:
            array: rightscale.Resource array object
        """
        params = self._generate_rightscale_params(
            'server_array', {'state': 'disabled'})

        if self._dry:
            self.log.info('Would have updated array "%s" with params: %s' %
                          (array.soul['name'], params))
            raise gen.Return()

        self.log.info('Disabling Array "%s"' % array.soul['name'])
        yield self._client.update(array, params)
        raise gen.Return()

    @gen.coroutine
    def _execute(self):
        # First, find the array we're going to be terminating.
        arrays = yield self._find_server_arrays(self.option('array'),
                                                raise_on=self._raise_on,
                                                allow_mock=self._allow_mock,
                                                exact=self.option('exact'))

        # Disable the array so that no new instances launch. Ignore the result
        # of this opertaion -- as long as it succeeds, we're happy. No need to
        # store the returned server array object.
        yield self._apply(self._disable_array, arrays)

        # Optionally terminate all of the instances in the array first.
        yield self._apply(self._terminate_all_instances, arrays)

        # Wait...
        yield self._apply(self._wait_until_empty, arrays)

        raise gen.Return()


class Destroy(Terminate):

    """Destroy a ServerArray in RightScale

    Destroys a ServerArray in RightScale by first invoking the Terminate actor,
    and then deleting the array as soon as all of the running instances have
    been terminated.

    **Options**

    :array:
      (str) The name of the ServerArray to destroy

    :exact:
      (bool) Whether or not to search for the exact array name.
      (default: `true`)

    :strict:
      (bool) Whether or not to fail if the ServerArray does not exist.
      (default: `true`)

    **Examples**

    .. code-block:: json

       { "desc": "Destroy my array",
         "actor": "rightscale.server_array.Destroy",
         "options": {
           "array": "my-array"
         }
       }

    .. code-block:: json

       { "desc": "Destroy many arrays",
         "actor": "rightscale.server_array.Destroy",
         "options": {
           "array": "array-prefix",
           "exact": false,
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* search for the `array`, but allows it to be
    missing because its highly likely that the array does not exist yet. If the
    array does not exist, a mocked array object is created for the rest of the
    execution.

    During the rest of the execution, the code bypasses making any real changes
    and just tells you what changes it would have made.

    Example *dry* output::

       [Destroy Test (DRY Mode)] Beginning
       [Destroy Test (DRY Mode)] Terminating array before destroying it.
       [Destroy Test (terminate) (DRY Mode)] Array "my-array" not found --
       creating a mock.
       [Destroy Test (terminate) (DRY Mode)] Disabling Array "my-array"
       [Destroy Test (terminate) (DRY Mode)] Would have terminated all array
       "<mocked array my-array>" instances.
       [Destroy Test (terminate) (DRY Mode)] Pretending that array <mocked
       array my-array> instances are terminated.
       [Destroy Test (DRY Mode)] Pretending to destroy array "<mocked array
       my-array>"
       [Destroy Test (DRY Mode)] Finished successfully. Result: True
    """

    @gen.coroutine
    def _destroy_array(self, array):
        """
        TODO: Handle exceptions if the array is not terminatable.
        """
        if self._dry:
            self.log.info('Pretending to destroy array "%s"' %
                          array.soul['name'])
            raise gen.Return()

        self.log.info('Destroying array "%s"' % array.soul['name'])
        yield self._client.destroy_server_array(array)
        raise gen.Return()

    @gen.coroutine
    def _execute(self):
        # Call the Terminate _execute function first
        yield super(Destroy, self)._execute()

        # Find the array we're going to be destroying.
        arrays = yield self._find_server_arrays(self.option('array'),
                                                raise_on=self._raise_on,
                                                allow_mock=self._allow_mock,
                                                exact=self.option('exact'))
        yield self._apply(self._destroy_array, arrays)
        raise gen.Return()


class Launch(ServerArrayBaseActor):

    """Launch instances in a ServerArray

    Launches instances in an existing ServerArray and waits until that array
    has become healthy before returning. *Healthy* means that the array has at
    least the user-specified ``count`` or ``max_count`` number of instances
    running as defined by the array definition in RightScale.

    **Options**

    :array:
      (str) The name of the ServerArray to launch
    :count:
      (str, int) Optional number of instance to launch. Defaults to max_count
      of the array.
    :success_pct':
      (str, int) Optional percent (0-100) to wait for instances to launch
      before exiting this actor as successful. Default: 100.

    :enable:
      (bool) Should the autoscaling of the array be enabled? Settings this to
      `false`, or omitting the parameter will not disable an enabled array.

    :exact:
      (bool) Whether or not to search for the exact array name.
      (default: `true`)

    **Examples**

    .. code-block:: json

       { "desc": "Enable array and launch it",
         "actor": "rightscale.server_array.Launch",
         "options": {
           "array": "my-array",
           "enable": true
         }
       }

    .. code-block:: json

       { "desc": "Enable arrays starting with my-array and launch them",
         "actor": "rightscale.server_array.Launch",
         "options": {
           "array": "my-array",
           "enable": true,
           "exact": false
         }
       }

    .. code-block:: json

       { "desc": "Enable array and launch 1 instance",
         "actor": "rightscale.server_array.Launch",
         "options": {
           "array": "my-array",
           "count": 1
         }
       }

    **Dry Mode**

    In Dry mode this actor *does* search for the ``array``, but allows it to be
    missing because its highly likely that the array does not exist yet. If the
    array does not exist, a mocked array object is created for the rest of the
    execution.

    During the rest of the execution, the code bypasses making any real changes
    and just tells you what changes it would have made.

    Example *dry* output::

       [Launch Array Test #0 (DRY Mode)] Verifying that array "my-array" exists
       [Launch Array Test #0 (DRY Mode)] Array "my-array" not found -- creating
           a mock.
       [Launch Array Test #0 (DRY Mode)] Enabling Array "my-array"
       [Launch Array Test #0 (DRY Mode)] Launching Array "my-array" instances
       [Launch Array Test #0 (DRY Mode)] Would have launched instances of array
           <MagicMock name='my-array.self.show().soul.__getitem__()'
           id='4420453200'>
       [Launch Array Test #0 (DRY Mode)] Pretending that array <MagicMock
           name='my-array.self.show().soul.__getitem__()' id='4420453200'>
           instances are launched.
    """

    all_options = {
        'array': (str, REQUIRED, 'ServerArray name to launch'),
        'count': (
            (int, str), False,
            "Number of server to launch. Default: up to array's max count"),
        'success_pct': (
            (int, str), 100,
            '% of servers to wait for. Default: 100% of `count`.'),
        'enable': (bool, False, 'Enable autoscaling?'),
        'exact': (bool, True, (
            'Whether to search for multiple ServerArrays and act on them.')),
    }

    def __init__(self, *args, **kwargs):
        """Check Actor prerequisites."""

        # Base class does everything to set up a generic class
        super(Launch, self).__init__(*args, **kwargs)

        try:
            int(self._options.get('count', False))
        except ValueError:
            raise exceptions.InvalidOptions('`count` must be an integer.')

    @gen.coroutine
    def _wait_until_healthy(self, array, sleep=60):
        """Sleep until a server array has its max_count servers running.

        This loop monitors the server array for its current live instance count
        and waits until the count hits zero before progressing.

        TODO: Add a timeout setting.

        Args:
            array: rightscale.Resource array object
            sleep: Integer time to sleep between checks (def: 60)
        """
        if self._dry:
            self.log.info('Pretending that array %s instances are launched.'
                          % array.soul['name'])
            raise gen.Return()

        max_count = int(self.option('count'))
        success_pct = float(self.option('success_pct'))
        if not max_count:
            max_count = int(array.soul['elasticity_params']
                            ['bounds']['max_count'])

        enough_count = int(math.ceil(max_count * (success_pct / 100.0)))

        while True:
            instances = yield self._client.get_server_array_current_instances(
                array, filters=['state==operational'])
            count = len(instances)
            self.log.info('%s instances found, waiting for %s/%s' %
                          (count, enough_count, max_count))

            if count >= enough_count:
                raise gen.Return()

            self.log.debug('Sleeping..')
            yield gen.sleep(sleep)

    @gen.coroutine
    def _launch_instances(self, array, count=False):
        """Launch new instances in a specified array.

        Instructs RightScale to launch instances, specified amount, or array's
        autoscaling 'max' value, in a syncronous or async way.

        TODO: Ensure that if 'count' is supplied, its *added* to the current
        array 'server instance count'. This allows the actor to launch 10 new
        servers in an already existing array, and wait until all 10 + the
        original group of servers are Operational.

        Args:
            array - rightscale ServerArray object
            count - `False` to use array's _max_ value
                    `int` to launch a specific number of instances
        """
        if not count:
            # Get the current max_count setting from the ServerArray object
            max_count = int(
                array.soul['elasticity_params']['bounds']['max_count'])

            instances = yield self._client.get_server_array_current_instances(
                array, filters=['state==operational'])
            current_count = len(instances)

            # Launch *up to* max_count. Not *new* max_count.
            count = max_count - current_count

            # Silly sanity check. If count < 0, set it to 0. There is no
            # concept of launching "negative" instance counts.
            if count < 0:
                count = 0

        if self._dry:
            self.log.info('Would have launched %s instances of array %s' % (
                          count, array.soul['name']))
            raise gen.Return()

        if count < 1:
            self.log.warning((
                'This array already has %s instances, and '
                'max_count is set to %s') % (current_count, max_count))
            raise gen.Return()

        self.log.info('Launching %s instances of array %s' % (
                      count, array.soul['name']))

        # Launch!
        yield self._client.launch_server_array(array, count=count)
        self.log.info('Launched %s instances for array %s' % (
                      count, array.soul['name']))

        raise gen.Return()

    @gen.coroutine
    def _enable_array(self, array):
        """Enable AutoScaling in a SeverArray.

        args:
            array: rightscale.Resource ServerArray Object
        """
        # This means that RightScale will auto-scale-up the array as soon as
        # their next scheduled auto-scale run hits (usually 60s). Store the
        # newly updated array.
        if self.option('enable'):
            if not self._dry:
                self.log.info('Enabling Array "%s"' % array.soul['name'])
                params = self._generate_rightscale_params(
                    'server_array', {'state': 'enabled'})
                array = yield self._client.update(array, params)
            else:
                self.log.info('Would enable array "%s"' % array.soul['name'])

    @gen.coroutine
    def _execute(self):
        # First, find the array we're going to be launching...
        arrays = yield self._find_server_arrays(
            self.option('array'),
            exact=self.option('exact'))

        # Enable the array, then launch it
        yield self._apply(self._enable_array, arrays)
        yield self._apply(self._launch_instances, arrays,
                          int(self.option('count')))

        # Now, wait until the number of healthy instances in the array matches
        # the max_count (or is greater than) of that array.
        yield self._apply(self._wait_until_healthy, arrays)
        raise gen.Return()


class Execute(ServerArrayBaseActor):

    """Executes a RightScale script/recipe on a ServerArray

    Executes a RightScript or Recipe on a set of hosts in a ServerArray in
    RightScale using individual calls to the live running instances. These can
    be found in your RightScale account under *Design -> RightScript* or
    *Design -> Cookbooks*

    The RightScale API offers a *multi_run_executable* method that can be used
    to run a single script on all servers in an array -- but unfortunately this
    API method provides no way to monitor the progress of the individual jobs
    on the hosts. Furthermore, the method often executes on recently terminated
    or terminating hosts, which throws false-negative error results.

    Our actor explicitly retrieves a list of the *operational* hosts in an
    array and kicks off individual execution tasks for every host. It then
    tracks the execution of those tasks from start to finish and returns the
    results.

    **Options**

    :array:
      (str) The name of the ServerArray to operate on

    :script:
      (str) The name of the RightScript or Recipe to execute

    :expected_runtime:
      (str, int) Expected number of seconds to execute.
      (default: `5`)

    :concurrency:
      Max number of concurrent executions. This will fire off N executions
      in parallel, and continue with the remained as soon as the first
      execution is done. This is faster than creating N Sync executions.
      **Note**: When applied to multiple (M) arrays cumulative concurrency
      accross all arrays will remain at N. It will not be M x N.

    :inputs:
      (dict) Dictionary of Key/Value pairs to use as inputs for the script

    :exact:
      (str) Boolean whether or not to search for the exact array name.
      (default: `true`)

    **Examples**

    .. code-block:: json

        { "desc":" Execute script on my-array",
          "actor": "rightscale.server_array.Execute",
          "options": {
            "array": "my-array",
            "script": "connect to elb",
            "expected_runtime": 3,
            "inputs": {
              "ELB_NAME": "text:my-elb"
            }
          }
        }

    **Dry Mode**

    In Dry mode this actor *does* search for the `array`, but allows it to be
    missing because its highly likely that the array does not exist yet. If the
    array does not exist, a mocked array object is created for the rest of the
    execution.

    During the rest of the execution, the code bypasses making any real changes
    and just tells you what changes it would have made.

    Example *dry* output::

        [Destroy Test (DRY Mode)] Verifying that array "my-array" exists
        [Execute Test (DRY Mode)]
            kingpin.actors.rightscale.server_array.Execute Initialized
        [Execute Test (DRY Mode)] Beginning execution
        [Execute Test (DRY Mode)] Verifying that array "my-array" exists
        [Execute Test (DRY Mode)] Would have executed "Connect instance to ELB"
            with inputs "{'inputs[ELB_NAME]': 'text:my-elb'}" on "my-array".
        [Execute Test (DRY Mode)] Returning result: True
    """

    all_options = {
        'array': (str, REQUIRED,
                  'ServerArray name on which to execute a script.'),
        'exact': (bool, True, (
            'Whether to search for multiple ServerArrays and act on them.')),
        'script': (str, REQUIRED,
                   'RightScale RightScript or Recipe to execute.'),
        'expected_runtime': (int, 5, 'Expected number of seconds to execute.'),
        'concurrency': (int, 0, "Max number of concurrent executions."),
        'inputs': (dict, {}, (
            'Inputs needed by the script. Read _generate_rightscale_params.'))
    }

    @gen.coroutine
    def _get_operational_instances(self, array):
        """Gets a list of Operational instances and returns it.

        Warns on any non-Operational instances to let the operator know that
        their script may not execute there.

        Args:
            array: rightscale.Resource ServerArray Object
        """
        # Get all non-terminated instances
        all_instances = yield self._client.get_server_array_current_instances(
            array, filters=['state<>terminated'])

        # Filter out the Operational ones from the Non-Operational (booting,
        # etc) instances.
        op = [inst for inst in all_instances if inst.soul['state'] ==
              'operational']
        non_op = [inst for inst in all_instances if inst.soul['state'] !=
                  'operational']

        # Warn that there are Non-Operational instances and move on.
        non_op_count = len(non_op)
        if non_op_count > 0:
            self.log.warning(
                'Found %s instances (in %s) in a non-Operational state, '
                'will not execute on these hosts!' %
                (non_op_count, array.soul['name']))

        self.log.info('Found %s instances (in %s) in the Operational state.' %
                      (len(op), array.soul['name']))
        raise gen.Return(op)

    @gen.coroutine
    def _check_script(self, script_name):
        if '::' in script_name:
            script = yield self._client.find_cookbook(script_name)
        else:
            script = yield self._client.find_right_script(script_name)

        raise gen.Return(bool(script))

    def _check_inputs(self):
        """Check that rightscale inputs are formatted properly.

        For more information read:
        http://reference.rightscale.com/api1.5/resources/ResourceInputs.html

        Raises:
            InvalidOptions
        """
        inputs = self.option('inputs')
        issues = False
        types = ('text', 'ignore', 'env', 'cred', 'key', 'array')
        for key, value in list(inputs.items()):
            if value.split(':')[0] not in types:
                issues = True
                self.log.error('Value for %s needs to begin with %s'
                               % (key, types))

        if issues:
            raise exceptions.InvalidOptions('One or more inputs has a problem')

    @gen.coroutine
    def _wait_for_all_tasks(self, task_pairs):
        """Wait for all instances to succeed, or print audit entry if failed.

        Args:
            task_pairs: list of tuples produced by run_executable_on_instances
                [(instance, task), (instance, task)]

        Returns:
            boolean: All tasks succeeded. If at least 1 failed - this is False
        """

        task_count = len(task_pairs)
        self.log.info('Queueing %s tasks' % task_count)
        task_waiting = []

        for instance, task in task_pairs:
            task_name = 'Executing "%s" on instance: %s' % (
                self.option('script'), instance.soul['name'])

            task_waiting.append(self._client.wait_for_task(
                task=task,
                task_name=task_name,
                sleep=self.option('expected_runtime'),
                loc_log=self.log,
                instance=instance
            ))

        self.log.info('Waiting for %s tasks to finish...' % task_count)
        statuses = yield task_waiting

        raise gen.Return(all(statuses))

    @gen.coroutine
    def _exec_and_wait(self, name, inputs, instance, sleep=5):
        """Start execution and wait for completion on a single instance.

        This shim combines the api calls for run_executable_on_instances and
        wait_for_task.

        Args:
            name: Recipe or RightScript String Name
            inputs: Dict of Key/Value Input Pairs
            instance: a single instance object of rightscale.Resource
            sleep: number of seconds to wait before the first status check

        Returns:
            success of wait_for_task()
        """
        tasks = yield self._client.run_executable_on_instances(
            name, inputs, instances=[instance])

        # tasks[0][1] is because there's only 1 task and
        # run_executable_on_instances returns (instance, task) tuple
        success = yield self._client.wait_for_task(
            task=tasks[0][1], task_name=name, sleep=sleep, loc_log=self.log,
            instance=instance)

        raise gen.Return(success)

    @gen.coroutine
    def _execute_array_with_concurrency(self, arrays, inputs):
        """Executes a script on many arrays with limited instance-concurrency.

        This method leverages the same rightscale.api methods as the non
        concurrent method, with the exception that it waits for tasks to finish
        after queuing the limit. The method has to know when a task is complete
        in order to cleverly schedule the next task.

        args:
            arrays: A list of, or a single instance of rightscale.Resource
                    ServerArray objects
            inputs: A string of inputs generated by
                    self._generate_rightscale_params()
        """
        if not isinstance(arrays, list):
            arrays = [arrays]

        instances = []
        for array in arrays:
            new_inst = yield self._get_operational_instances(array)
            instances.extend(new_inst)

        count = len(instances)

        if self._dry:
            self.log.info((
                'Would have executed "%s" with inputs "%s" on %s instances '
                'on %s arrays with limited concurrency of %s.') % (
                    self.option('script'), inputs, count, len(arrays),
                    self.option('concurrency')))
            raise gen.Return()

        self.log.info('Concurrency set to %s' % self.option('concurrency'))
        tasks = []
        for i in instances:
            tasks.append(self._exec_and_wait(
                name=self.option('script'),
                inputs=inputs,
                instance=i,
                sleep=self.option('expected_runtime')))

            running_tasks = len([t for t in tasks if not t.done()])
            if running_tasks < self.option('concurrency'):
                # We can queue more tasks, continue the loop to add one more.
                continue

            self.log.debug('Concurrency saturated. Waiting...')
            while running_tasks >= self.option('concurrency'):
                yield gen.moment
                running_tasks = len([t for t in tasks if not t.done()])

            self.log.debug('Concurrency desaturated: %s<%s. Continuing.' % (
                running_tasks, self.option('concurrency')))

        statuses = yield tasks
        raise gen.Return(all(statuses))

    @gen.coroutine
    def _execute_array(self, array, inputs):
        """Executes a script on an array.

        This method does the real work. It gets a list of instances from an
        array, finds a script, executes the script, and then waits for the
        results of the script. Ultimately it raises a failure if the script
        fails, or simply exits cleanly.

        Note: This is separated out from the _execute() method to facilitate
        using the self._apply() function with multiple arrays.

        args:
            array: rightscale.Resource ServerArray
            inputs: A string of inputs generated by
                    self._generate_rightscale_params()
        """
        instances = yield self._get_operational_instances(array)

        if self._dry:
            self.log.info(
                'Would have executed "%s" with inputs "%s" on "%s".'
                % (self.option('script'), inputs, array.soul['name']))
            raise gen.Return()

        count = len(instances)
        # Execute the script on all of the servers in the array and store the
        # task status resource records.
        self.log.info(
            'Executing "%s" on %s instances in the array "%s"' %
            (self.option('script'), count, array.soul['name']))

        try:
            task_pairs = yield self._client.run_executable_on_instances(
                self.option('script'), inputs, instances)
        except api.ServerArrayException as e:
            self.log.critical('Script execution error: %s' % e)
            raise exceptions.RecoverableActorFailure(
                'Invalid parameters supplied to execute script.')

        # Finally, monitor all of the tasks for completion.
        successful = yield self._wait_for_all_tasks(task_pairs)

        # If not all of the executions succeeded, raise an exception.
        if not successful:
            self.log.critical('One or more tasks failed.')
            raise TaskExecutionFailed()
        else:
            self.log.info('Completed %s tasks.' % count)

        raise gen.Return()

    @gen.coroutine
    def _execute(self):
        """Executes the actor.

        Logs into RightScale, validates (if in dry run) that the script
        actually exists, and then executes the script on all of the matched
        server arrays.
        """
        # Munge our inputs into something that RightScale likes
        inputs = self._generate_rightscale_params(
            'inputs', self.option('inputs'))

        # Theres no way to 'test' the actual execution of the rightscale
        # scripts, so we'll just check that it exists.
        if self._dry:
            script_found = yield self._check_script(self.option('script'))

            if not script_found:
                msg = 'Script "%s" not found!' % self.option('script')
                raise exceptions.InvalidOptions(msg)

            self._check_inputs()

        # First, find the array we're going to be launching. Get a list back of
        # the 'operational' instances that we are able to execute scripts
        # against.
        arrays = yield self._find_server_arrays(
            self.option('array'), exact=self.option('exact'))
        if self.option('concurrency'):
            yield self._execute_array_with_concurrency(arrays, inputs)
        else:
            yield self._apply(self._execute_array, arrays, inputs)



"""Tests for the actors.slack package"""

import mock

from tornado import testing

from kingpin.actors import slack
from kingpin.actors import exceptions
from kingpin.actors.test.helper import mock_tornado
import importlib


__author__ = 'Matt Wise <matt@nextdoor.com>'


class TestSlackBase(testing.AsyncTestCase):

    """Unit tests for the Slack Base actor."""

    def setUp(self, *args, **kwargs):
        # For most tests, mock out the TOKEN
        super(TestSlackBase, self).setUp(*args, **kwargs)
        slack.TOKEN = 'Unittest'

    def test_init(self):
        actor = slack.SlackBase('Unit test action', {})

        # Ensure that the actor._slack_client is configured with a dictionary
        # that contains the token in it.
        slack_client_tokens = actor._slack_client._client._tokens
        self.assertEqual(
            slack_client_tokens,
            {'token': 'Unittest'})

    def test_init_missing_creds(self):
        # Un-set the token now and make sure the init fails
        slack.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            slack.SlackBase('Unit Test Action', {})
        # Reload the slack library to re-get the token
        importlib.reload(slack)

    def test_check_results_with_ok_results(self):
        actor = slack.SlackBase('Unit test action', {})
        results = {
            "ok": True, "channel": "C03H4GRDF", "ts": "1423092527.000006",
            "message": {
                "text": "Hi, testing!",
                "username": "Kingpin",
                "type": "message",
                "subtype": "bot_message",
                "ts": "1423092527.000006"
            }
        }
        self.assertEqual(None, actor._check_results(results))

    def test_check_results_with_invalid_creds(self):
        actor = slack.SlackBase('Unit test action', {})
        results = {'ok': False, 'error': 'invalid_auth'}
        with self.assertRaises(exceptions.InvalidCredentials):
            actor._check_results(results)

    def test_check_results_with_unexpected_results(self):
        actor = slack.SlackBase('Unit test action', {})
        results = 'got some unexpected result'
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            actor._check_results(results)


class TestMessage(testing.AsyncTestCase):

    """Unit tests for the Slack Message actor."""

    def setUp(self, *args, **kwargs):
        # For most cases, mock out the TOKEN
        super(TestMessage, self).setUp(*args, **kwargs)
        slack.TOKEN = 'Unittest'

        self.actor = slack.Message(
            'Unit test message',
            {'channel': '#testing',
             'message': 'Unittest'})
        self._slack_mock = mock.MagicMock(name='SlackAPIMock')
        self.actor._slack_client = self._slack_mock

    @testing.gen_test
    def test_execute_dry(self):
        # Mock out the calls to SlackAPI.auth_test().auth_test()
        auth_test_mock = mock.MagicMock(name='auth_test')
        auth_test_mock.http_post.side_effect = mock_tornado({'ok': 'true'})
        self._slack_mock.auth_test.return_value = auth_test_mock

        # Ensure we're dry
        self.actor._dry = True
        ret = yield self.actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        auth_test_mock.http_post.assert_has_calls([mock.call()])

    @testing.gen_test
    def test_execute(self):
        # Mock out the calls to SlackAPI.chat_postMessage().http_post()
        post_mock = mock.MagicMock()
        post_mock.http_post.side_effect = mock_tornado({'ok': 'true'})
        self._slack_mock.chat_postMessage.return_value = post_mock

        ret = yield self.actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        post_mock.http_post.assert_has_calls([mock.call(
            username='Kingpin', unfurl_links=True, text='Unittest',
            unfurl_media=True, parse='none', link_names=1, channel='#testing'
        )])

    @testing.gen_test
    def test_execute_list_rooms(self):
        actor = slack.Message(
            'Unit test message',
            {'channel': ['#testing', '#testing2'],
             'message': 'Unittest'})
        actor._slack_client = self._slack_mock

        # Mock out the calls to SlackAPI.chat_postMessage().http_post()
        post_mock = mock.MagicMock()
        post_mock.http_post.side_effect = mock_tornado({'ok': 'true'})
        self._slack_mock.chat_postMessage.return_value = post_mock

        ret = yield actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        post_mock.http_post.assert_has_calls([mock.call(
            username='Kingpin', unfurl_links=True, text='Unittest',
            unfurl_media=True, parse='none', link_names=1, channel='#testing'
        )])

    @testing.gen_test
    def test_execute_csv_rooms(self):
        actor = slack.Message(
            'Unit test message',
            {'channel': '#testing, #testing2',
             'message': 'Unittest'})
        actor._slack_client = self._slack_mock

        # Mock out the calls to SlackAPI.chat_postMessage().http_post()
        post_mock = mock.MagicMock()
        post_mock.http_post.side_effect = mock_tornado({'ok': 'true'})
        self._slack_mock.chat_postMessage.return_value = post_mock

        ret = yield actor._execute()
        self.assertEqual(None, ret)

        # Ensure the calls were made to the API
        post_mock.http_post.assert_has_calls([mock.call(
            username='Kingpin', unfurl_links=True, text='Unittest',
            unfurl_media=True, parse='none', link_names=1, channel='#testing'
        )])



"""Tests for the actors.rollbar package"""

from nose.plugins.attrib import attr

from tornado import testing

from kingpin import version
from kingpin.actors import rollbar
from kingpin.actors import exceptions
import importlib


__author__ = 'Matt Wise <matt@nextdoor.com>'


class IntegrationRollbarDeploy(testing.AsyncTestCase):

    """Simple high level integration tests agains the Rollbar API.

    These tests actually hit the Rollbar API and test that the code
    works, as well as validate that the API token is working properly.

    Require ROLLBAR_TOKEN environment variable to be set.
    """

    integration = True

    @attr('rollbar', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_1a_init_without_environment_creds(self):
        # Un-set the token now and make sure the init fails
        rollbar.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            rollbar.Deploy(
                'Unit Test Action',
                {'environment': 'kingpin-integration-testing',
                 'revision': version.__version__,
                 'local_username': 'Kingpin Integration Testing',
                 'comment': 'Integration Tests Are Good, MmmKay'})

        # Reload the rollbar package so it gets our environment variable back.
        importlib.reload(rollbar)

    @attr('rollbar', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_test_2a_execute_with_invalid_creds(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'kingpin-integration-testing',
             'revision': version.__version__,
             'local_username': 'Kingpin Integration Testing',
             'comment': 'Integration Tests Are Good, MmmKay'}, dry=True)

        # Valid response test
        actor._token = 'Invalid'
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()

    @attr('rollbar', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_test_2b_execute_dry(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'kingpin-integration-testing',
             'revision': version.__version__,
             'local_username': 'Kingpin Integration Testing',
             'comment': 'This should never appear in Rollbar'}, dry=True)
        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('rollbar', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_2c_execute_real(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'kingpin-integration-testing',
             'revision': version.__version__,
             'local_username': 'Kingpin Integration Testing',
             'comment': 'Integration Tests Are Good, MmmKay'})

        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('rollbar', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_2d_execute_real_with_rollbar_username(self):
        actor = rollbar.Deploy(
            'Unit Test Action',
            {'environment': 'kingpin-integration-testing',
             'revision': version.__version__,
             'local_username': 'Kingpin Integration Testing',
             'rollbar_username': 'Kingpin Integration Username',
             'comment': 'Now, with a rollbar_username too!'})

        res = yield actor.execute()
        self.assertEqual(res, None)



import logging

from tornado import httpclient
from tornado import testing
import mock

from kingpin import exceptions as kingpin_exceptions
from kingpin.actors import exceptions
from kingpin.actors import misc
from kingpin.actors.test.helper import mock_tornado
import importlib

log = logging.getLogger(__name__)


class TestNote(testing.AsyncTestCase):

    @testing.gen_test
    def test_log(self):
        note = misc.Note('Test', {'message': 'Hello World'})
        note.log = mock.Mock()
        yield note._execute()
        self.assertEqual(note.log.info.call_count, 1)


class TestMacro(testing.AsyncTestCase):

    def setUp(self):
        super(TestMacro, self).setUp()
        importlib.reload(misc)

    def test_init(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock(return_value='unit-test-macro')

        with mock.patch('kingpin.utils.convert_script_to_dict') as j2d, \
                mock.patch('kingpin.schema.validate') as schema_validate, \
                mock.patch('kingpin.actors.utils.get_actor') as get_actor:

            j2d.return_value = {
                'desc': 'unit test',
                'actor': 'unit test',
                'options': {}
            }

            actor = misc.Macro('Unit Test', {'macro': 'test.json',
                                             'tokens': {}},
                               init_tokens={})

            j2d.assert_called_with(script_file='unit-test-macro', tokens={})
            self.assertEqual(schema_validate.call_count, 1)
            self.assertEqual(actor.initial_actor, get_actor())

    def test_init_nested_tokens(self):
        # Quick test to ensure that a series of nested groups and macro actors
        # will pass supplied tokens all the way down to their sub actors.

        # Generate a single outer actor. This actor will create many internal
        # actors.
        init_tokens = {'SLEEP': 0}
        actor = misc.Macro(
            options={'macro': 'examples/misc.macro/outer_group.yaml'},
            init_tokens=init_tokens)

        # Ensure that the initial misc.Macro actor, and the initial actor it
        # created (from outer_group.yaml) both have the appropriate init
        # tokens. This helps ensure that we used .copy() properly a well as the
        # fact that the tokens were passed down appropriately.
        self.assertEqual(actor._init_tokens, init_tokens)
        self.assertEqual(actor.initial_actor._init_tokens, init_tokens)

        # Ensure that the nested misc.Macro actor from outer_macro.yaml got
        # init_tokens, AND the 'FOO' token from outer_group.yaml's own
        # definition.
        self.assertEqual(actor.initial_actor._actions[0]._init_tokens,
                         {'SLEEP': 0, 'FOO': 'weee'})

        # Next ensure that the mostly nested examples/misc.macro/inner.yaml
        # actor got the SLEEP, FOO, and DESC tokens.
        self.assertEqual(
            actor.initial_actor._actions[0].initial_actor._init_tokens,
            {'SLEEP': 0, 'FOO': 'weee', 'DESC': 'Sleeping for a while'})

        # Finally, ensure the super nested
        s = actor.initial_actor._actions[0].initial_actor.initial_actor
        self.assertEqual(
            s._init_tokens,
            {'SLEEP': 0, 'FOO': 'weee', 'DESC': 'Sleeping for a while'})

    def test_init_group(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock(return_value='unit-test-macro')

        with mock.patch('kingpin.utils.convert_script_to_dict') as j2d, \
                mock.patch('kingpin.schema.validate') as schema_validate, \
                mock.patch('kingpin.actors.group.Sync') as sync_actor:

            j2d.return_value = [{
                'desc': 'unit test',
                'actor': 'unit test',
                'options': {}
            }]

            actor = misc.Macro('Unit Test', {'macro': 'test.json'})

            j2d.assert_called_with(script_file='unit-test-macro', tokens={})
            self.assertEqual(schema_validate.call_count, 1)
            self.assertEqual(actor.initial_actor, sync_actor())

    def test_init_remote(self):
        misc.Macro._get_config_from_script = mock.Mock()
        misc.Macro._get_config_from_script.return_value = {}
        misc.Macro._check_schema = mock.Mock()
        with mock.patch('kingpin.actors.utils.get_actor'):
            with mock.patch.object(httpclient.HTTPClient, 'fetch') as fetch:
                fetch.return_value.body = "foo"
                misc.Macro('Unit Test', {'macro': 'http://test.json',
                                         'tokens': {}})

    def test_init_dry(self):
        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_config_from_script = mock.Mock()
        misc.Macro._get_config_from_script.return_value = {}
        misc.Macro._check_schema = mock.Mock()

        with mock.patch('kingpin.utils.convert_script_to_dict') as j2d, \
                mock.patch('kingpin.schema.validate'), \
                mock.patch('kingpin.actors.utils.get_actor'):
            j2d.return_value = {
                'desc': 'unit test',
                'actor': 'unit test',
                'options': {}
            }

            actor = misc.Macro('Unit Test',
                               {'macro': 'examples/test/sleep.json',
                                'tokens': {}},
                               dry=True)

            self.assertTrue(actor.initial_actor._dry)

    def test_init_with_errors(self):

        # Remote files are prohibited for now
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Unit Test', {'macro': 'ftp://fail.test.json',
                                     'tokens': {}})

        # Remote file with bad URL
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Unit Test', {'macro': 'http://fail.test.json',
                                     'tokens': {}})

        # Non-existent file
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            misc.Macro('Unit Test', {'macro': 'dontcreatethis.json',
                                     'tokens': {}})

        # We don't want the rest of the tests failing on downloading this file.
        misc.Macro._get_macro = mock.Mock(return_value='unit-test-file')

        # Schema failure
        with mock.patch('kingpin.utils.convert_script_to_dict') as j2d:
            j2d.return_value = {
                'desc': 'unit test',
                'options': {}  # `actor` keyword is missing
            }

            with self.assertRaises(exceptions.UnrecoverableActorFailure):
                misc.Macro('Unit Test', {'macro': 'test.json',
                                         'tokens': {}})

        # JSON syntax error
        with mock.patch('kingpin.utils.convert_script_to_dict') as j2d:

            j2d.side_effect = kingpin_exceptions.InvalidScript('Fail!')

            with self.assertRaises(exceptions.UnrecoverableActorFailure):
                misc.Macro('Unit Test', {'macro': 'test.json',
                                         'tokens': {}})

    @testing.gen_test
    def test_execute(self):

        misc.Macro._check_macro = mock.Mock()
        misc.Macro._get_macro = mock.Mock()
        misc.Macro._get_config_from_script = mock.Mock()
        misc.Macro._get_config_from_script.return_value = {}
        misc.Macro._check_schema = mock.Mock()

        with mock.patch('kingpin.actors.utils.get_actor') as get_actor:
            actor = misc.Macro('Unit Test', {'macro': 'test.json',
                                             'tokens': {}},
                               dry=True)

            get_actor().execute = mock_tornado()
            yield actor._execute()

            self.assertEqual(get_actor().execute._call_count, 1)

    @testing.gen_test
    def test_orgchart(self):

        misc.Macro._get_macro = mock.Mock(name='unittestmacro')
        misc.Macro._get_config_from_script = mock.Mock(
            return_value=[{'actor': 'misc.Sleep', 'options': {'sleep': 0}}]
        )
        actor = misc.Macro('Unit test', {'macro': 'test'})

        self.assertEqual(len(actor.get_orgchart()), 3)  # Macro, Group, Sleep
        self.assertEqual(type(actor.get_orgchart()[0]), dict)


class TestSleep(testing.AsyncTestCase):

    @testing.gen_test
    def test_execute(self):
        # Call the executor and test it out
        actor = misc.Sleep('Unit Test Action', {'sleep': 0.1})
        res = yield actor.execute()

        # Make sure we fired off an alert.
        self.assertEqual(res, None)

    @testing.gen_test
    def test_execute_with_str(self):
        # Call the executor and test it out
        actor = misc.Sleep('Unit Test Action', {'sleep': '0.5'})
        res = yield actor.execute()

        # Make sure we fired off an alert.
        self.assertEqual(res, None)


class TestGenericHTTP(testing.AsyncTestCase):

    @testing.gen_test
    def test_execute_dry(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com'},
                                 dry=True)

        actor._fetch = mock_tornado()

        yield actor.execute()

        self.assertEqual(actor._fetch._call_count, 0)

    @testing.gen_test
    def test_execute(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com'})
        actor._fetch = mock_tornado({'success': {'code': 200}})

        yield actor.execute()

    @testing.gen_test
    def test_execute_data_json(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com',
                                  'data-json': {'foo': 'bar'}})
        actor._fetch = mock_tornado({'success': {'code': 200}})

        yield actor.execute()

    @testing.gen_test
    def test_execute_fail(self):
        actor = misc.GenericHTTP('Unit Test Action',
                                 {'url': 'http://example.com'})
        error = httpclient.HTTPError(code=401, response={})
        actor._fetch = mock_tornado(exc=error)

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()



import logging
import mock

from tornado import gen
from tornado import testing

from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors import utils
from kingpin.actors import misc
from kingpin.actors.test import helper


log = logging.getLogger(__name__)


class FakeActor(base.BaseActor):

    """Fake Actor use for Unit Tests"""

    def __init__(self, *args, **kwargs):
        super(FakeActor, self).__init__(*args, **kwargs)
        self.conn = mock.MagicMock()
        self.conn.call.return_value = helper.tornado_value(None)
        self.conn.call.__name__ = 'test_call'

    @gen.coroutine
    @utils.dry('Would have done {0}')
    def do_thing(self, thing):
        yield self.conn.call(thing)

    @gen.coroutine
    def execute(self):
        raise gen.Return(self.options['return_value'])


class TestUtils(testing.AsyncTestCase):

    def test_get_actor(self):
        actor_return_true = {
            'desc': 'returns true',
            'actor': 'kingpin.actors.test.test_utils.FakeActor',
            'options': {'return_value': True}}
        ret = utils.get_actor(actor_return_true, dry=True)
        self.assertEqual(True, ret._options['return_value'])
        self.assertEqual(FakeActor, type(ret))

    def test_get_actor_class(self):
        actor_string = 'misc.Sleep'
        ret = utils.get_actor_class(actor_string)
        self.assertEqual(type(misc.Sleep), type(ret))

    def test_get_actor_class_direct(self):
        actor_string = 'kingpin.actors.test.test_utils.FakeActor'
        ret = utils.get_actor_class(actor_string)
        self.assertEqual(type(FakeActor), type(ret))

    def test_get_actor_class_bogus_actor(self):
        actor_string = 'bogus.actor'
        with self.assertRaises(exceptions.InvalidActor):
            utils.get_actor_class(actor_string)

    @testing.gen_test
    def test_dry_decorator_with_dry_true(self):
        actor = FakeActor('Fake', options={}, dry=True)
        yield actor.do_thing('my thing string')
        self.assertFalse(actor.conn.called)

    @testing.gen_test
    def test_dry_decorator_with_dry_false(self):
        actor = FakeActor('Fake', options={}, dry=False)
        yield actor.do_thing('my thing string')
        actor.conn.call.assert_has_calls([mock.call('my thing string')])



import logging

from boto.exception import BotoServerError
from tornado import testing
import mock

from kingpin import utils
from kingpin.actors import exceptions
from kingpin.actors.aws import elb as elb_actor
from kingpin.actors.aws import settings
from kingpin.actors.test import helper
import importlib

log = logging.getLogger(__name__)


class TestRegisterInstance(testing.AsyncTestCase):

    def setUp(self):
        super(TestRegisterInstance, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(elb_actor)

    @testing.gen_test
    def test_add(self):
        act = elb_actor.RegisterInstance('UTA', {
            'elb': 'test',
            'region': 'us-east-1',
            'instances': 'test'})

        elb = mock.Mock()
        instance = 'i-un173s7'
        yield act._add(elb=elb, instances=[instance])

        elb.register_instances.assert_called_with([instance])

    @testing.gen_test
    def test_add_zones(self):
        act = elb_actor.RegisterInstance('UTA', {
            'elb': 'test',
            'region': 'us-east-1',
            'instances': 'test'})
        act.ec2_conn = mock.Mock()
        zone = mock.Mock()
        zone.name = 'unit-test-zone'
        act.ec2_conn.get_all_zones.return_value = [zone]

        elb = mock.Mock()
        elb.availability_zones = []

        yield act._check_elb_zones(elb=elb)

        elb.enable_zones.assert_called_with(set(['unit-test-zone']))

    @testing.gen_test
    def test_add_zones_noop(self):
        act = elb_actor.RegisterInstance('UTA', {
            'elb': 'test',
            'region': 'us-east-1',
            'instances': 'test'})
        act.ec2_conn = mock.Mock()
        zone = mock.Mock()
        zone.name = 'unit-test-zone'
        act.ec2_conn.get_all_zones.return_value = [zone]

        elb = mock.Mock()
        elb.availability_zones = ['unit-test-zone']

        yield act._check_elb_zones(elb=elb)

        self.assertEqual(elb.enable_zones.call_count, 0)

    @testing.gen_test
    def test_execute(self):
        act = elb_actor.RegisterInstance('UTA', {
            'elb': 'elb-test',
            'region': 'us-east-1',
            'instances': 'i-test'})

        act._find_elb = mock.Mock()
        act._find_elb.return_value = helper.tornado_value(mock.Mock())
        act._check_elb_zones = mock.Mock()
        act._check_elb_zones.return_value = helper.tornado_value(mock.Mock())
        act._add = mock.Mock()
        act._add.return_value = helper.tornado_value(mock.Mock())
        yield act._execute()

        act._find_elb.assert_called_with('elb-test')
        lb = yield act._find_elb()
        act._add.assert_called_with(elb=lb, instances=['i-test'])

    @testing.gen_test
    def test_execute_self(self):
        # No instance id specified
        act = elb_actor.RegisterInstance('UTA', {
            'elb': 'elb-test',
            'region': 'us-east-1'})

        act._find_elb = mock.Mock()
        act._find_elb.return_value = helper.tornado_value(mock.Mock())
        act._check_elb_zones = mock.Mock()
        act._check_elb_zones.return_value = helper.tornado_value(mock.Mock())
        act._add = mock.Mock()
        act._add.return_value = helper.tornado_value(mock.Mock())
        act._get_meta_data = helper.mock_tornado('i-test')
        yield act._execute()

        act._find_elb.assert_called_with('elb-test')
        lb = yield act._find_elb()
        act._add.assert_called_with(elb=lb, instances=['i-test'])


class TestDeregisterInstance(testing.AsyncTestCase):

    def setUp(self):
        super(TestDeregisterInstance, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(elb_actor)

    @testing.gen_test
    def test_remove(self):
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': 'test',
            'region': 'us-east-1',
            'instances': 'test'})

        elb = mock.Mock()
        instance = 'i-un173s7'

        act._wait_on_draining = mock.Mock()
        act._wait_on_draining.return_value = helper.tornado_value(mock.Mock())
        yield act._remove(elb=elb, instances=[instance])

        elb.deregister_instances.assert_called_with([instance])

    @testing.gen_test
    def test_wait_on_draining(self):
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': 'test',
            'region': 'us-east-1',
            'isntances': 'test'})

        # Quick test with draining enabled
        fake_elb_attrs = mock.Mock(name='attrs')
        fake_elb_attrs.connection_draining.enabled = True
        fake_elb_attrs.connection_draining.timeout = 0.1
        elb = mock.Mock()
        elb.get_attributes.return_value = fake_elb_attrs
        yield act._wait_on_draining(elb)
        elb.get_attributes.assert_called_with()

        # Quick re-test with draining disabled
        act._options['wait_on_draining'] = False
        elb.reset_mock()
        yield act._wait_on_draining(elb)
        elb.get_attributes.assert_not_called()

        # Quick re-test with draining disabled
        act._options['wait_on_draining'] = True
        fake_elb_attrs.connection_draining.enabled = False
        elb.reset_mock()
        yield act._wait_on_draining(elb)
        elb.get_attributes.assert_called_with()

    @testing.gen_test
    def test_execute(self):
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': 'elb-test',
            'region': 'us-east-1',
            'instances': 'i-test'})

        act._find_elb = mock.Mock()
        act._find_elb.return_value = helper.tornado_value(mock.Mock())
        act._remove = mock.Mock()
        act._remove.return_value = helper.tornado_value(mock.Mock())
        yield act._execute()

        act._find_elb.assert_called_with('elb-test')
        lb = yield act._find_elb()
        act._remove.assert_called_with(elb=lb, instances=['i-test'])

    @testing.gen_test
    def test_execute_wildcard(self):
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': '*',
            'region': 'us-east-1',
            'instances': 'i-test'})

        act._find_instance_elbs = mock.Mock()
        act._find_instance_elbs.return_value = helper.tornado_value(
            [mock.Mock()])
        act._remove = mock.Mock()
        act._remove.return_value = helper.tornado_value(mock.Mock())

        yield act._execute()

        act._find_instance_elbs.assert_called_with(['i-test'])

    @testing.gen_test
    def test_find_instance_elbs(self):
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': '*',
            'region': 'us-east-1',
            'instances': 'i-test'})

        fake_instance_1 = mock.Mock(name='i-1234567')
        fake_instance_1.id = 'i-1234567'
        fake_instance_2 = mock.Mock(name='i-test')
        fake_instance_2.id = 'i-test'

        fake_elb_1 = mock.Mock(name='elb_1')
        fake_elb_1.instances = [fake_instance_1]
        fake_elb_2 = mock.Mock(name='elb_2')
        fake_elb_2.instances = [fake_instance_1, fake_instance_2]
        fake_elbs = [fake_elb_1, fake_elb_2]

        act.elb_conn.get_all_load_balancers = mock.Mock()
        act.elb_conn.get_all_load_balancers.return_value = fake_elbs

        ret = yield act._find_instance_elbs(['i-test'])

        self.assertEqual(ret, [fake_elb_2])

    @testing.gen_test
    def test_execute_self(self):
        # No instance id specified
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': 'elb-test',
            'region': 'us-east-1'})

        act._find_elb = mock.Mock()
        act._find_elb.return_value = helper.tornado_value(mock.Mock())
        act._remove = mock.Mock()
        act._remove.return_value = helper.tornado_value(mock.Mock())
        act._get_meta_data = helper.mock_tornado('i-test')
        yield act._execute()

        act._find_elb.assert_called_with('elb-test')
        lb = yield act._find_elb()
        act._remove.assert_called_with(elb=lb, instances=['i-test'])

    @testing.gen_test
    def test_execute_dry(self):
        act = elb_actor.DeregisterInstance('UTA', {
            'elb': 'elb-test',
            'region': 'us-east-1',
            'instances': 'i-test'},
            dry=True)

        act._find_elb = mock.Mock()
        act._find_elb.return_value = helper.tornado_value(mock.Mock())
        yield act._execute()


class TestWaitUntilHealthy(testing.AsyncTestCase):

    def setUp(self):
        super(TestWaitUntilHealthy, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(elb_actor)

    @testing.gen_test
    def test_execute(self):

        actor = elb_actor.WaitUntilHealthy(
            'Unit Test Action', {'name': 'unit-test-queue',
                                 'region': 'us-west-2',
                                 'count': 3})

        actor._find_elb = mock.Mock(return_value=helper.tornado_value('ELB'))
        actor._is_healthy = mock.Mock(return_value=helper.tornado_value(True))

        val = yield actor._execute()
        self.assertEqual(actor._find_elb.call_count, 1)
        self.assertEqual(actor._is_healthy.call_count, 1)
        self.assertEqual(val, None)

    @testing.gen_test
    def test_execute_retry(self):

        actor = elb_actor.WaitUntilHealthy(
            'Unit Test Action', {'name': 'unit-test-queue',
                                 'region': 'us-west-2',
                                 'count': 3})

        actor._find_elb = mock.Mock(return_value=helper.tornado_value('ELB'))
        actor._is_healthy = mock.Mock(
            side_effect=[helper.tornado_value(False),
                         helper.tornado_value(True)])

        # Optional mock -- making the test quicker.
        short_sleep = utils.tornado_sleep(0)
        with mock.patch('kingpin.utils.tornado_sleep') as ts:
            ts.return_value = short_sleep
            val = yield actor._execute()

        self.assertEqual(actor._find_elb.call_count, 1)  # Don't refetch!
        self.assertEqual(actor._is_healthy.call_count, 2)  # Retry!
        self.assertEqual(val, None)

    @testing.gen_test
    def test_execute_dry(self):

        actor = elb_actor.WaitUntilHealthy(
            'Unit Test Action', {'name': 'unit-test-queue',
                                 'region': 'us-west-2',
                                 'count': 3},
            dry=True)

        actor._find_elb = mock.Mock(return_value=helper.tornado_value('ELB'))
        # NOTE: this is false, but assertion is True!
        actor._is_healthy = mock.Mock(return_value=helper.tornado_value(False))

        val = yield actor._execute()
        self.assertEqual(actor._find_elb.call_count, 1)
        self.assertEqual(actor._is_healthy.call_count, 1)
        self.assertEqual(val, None)

    @testing.gen_test
    def test_execute_fail(self):

        actor = elb_actor.WaitUntilHealthy(
            'Unit Test Action', {'name': 'unit-test-queue',
                                 'region': 'us-west-2',
                                 'count': 7})
        # ELB not found...
        actor.elb_conn.get_all_load_balancers = mock.Mock(
            side_effect=BotoServerError(400, 'LoadBalancerNotFound'))

        with self.assertRaises(elb_actor.base.ELBNotFound):
            yield actor.execute()

    def test_get_expected_count(self):
        actor = elb_actor.WaitUntilHealthy(
            'Unit Test Action', {'name': 'unit-test-queue',
                                 'region': 'us-west-2',
                                 'count': 3})

        self.assertEqual(actor._get_expected_count(5, 1), 5)
        self.assertEqual(actor._get_expected_count('50%', 20), 10)

    @testing.gen_test
    def test_is_healthy(self):
        actor = elb_actor.WaitUntilHealthy(
            'Unit Test Action', {'name': 'unit-test-queue',
                                 'region': 'us-west-2',
                                 'count': 3})

        elb = mock.Mock()
        elb.get_instance_health.return_value = [
            mock.Mock(state='InService'),
            mock.Mock(state='InService'),
            mock.Mock(state='InService'),
            mock.Mock(state='OutOfService'),
            mock.Mock(state='OutOfService'),
        ]
        val = yield actor._is_healthy(elb, 3)

        self.assertTrue(val)


class TestSetCert(testing.AsyncTestCase):

    def setUp(self):
        super(TestSetCert, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(elb_actor)

    @testing.gen_test
    def test_check_access(self):
        elb = mock.Mock()
        botoerror = BotoServerError('Fail', 'Unit test')
        botoerror.error_code = 'AccessDenied'
        elb.set_listener_SSL_certificate = mock.Mock(
            side_effect=botoerror)

        actor = elb_actor.SetCert(
            'Unit Test', {'name': 'unit-test',
                          'region': 'us-east-1',
                          'cert_name': 'unit-cert'}
        )

        # AccessDenied means check has failed.
        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._check_access(elb)

        # Anything else means the check has passed.
        botoerror.error_code = 'Cert Not Found'
        yield actor._check_access(elb)

    @testing.gen_test
    def test_get_cert_arn(self):
        cert = {
            'get_server_certificate_response': {
                'get_server_certificate_result': {
                    'server_certificate': {
                        'server_certificate_metadata': {
                            'arn': 'unit-test-arn-value'}}}}}

        actor = elb_actor.SetCert(
            'Unit Test', {'name': 'unit-test',
                          'region': 'us-east-1',
                          'cert_name': 'unit-cert'}
        )
        actor.iam_conn = mock.Mock()
        actor.iam_conn.get_server_certificate = mock.Mock(return_value=cert)

        arn = yield actor._get_cert_arn('test')

        self.assertEqual(actor.iam_conn.get_server_certificate.call_count, 1)
        self.assertEqual(arn, 'unit-test-arn-value')

        yield actor._get_cert_arn('test-new')
        # New name supplied, call count should be 2
        self.assertEqual(actor.iam_conn.get_server_certificate.call_count, 2)

    @testing.gen_test
    def test_get_cert_arn_fail(self):
        actor = elb_actor.SetCert(
            'Unit Test', {'name': 'unit-test',
                          'region': 'us-east-1',
                          'cert_name': 'unit-cert'}
        )

        actor.iam_conn = mock.Mock()
        error = BotoServerError(400, 'test')
        actor.iam_conn.get_server_certificate.side_effect = error

        with self.assertRaises(elb_actor.CertNotFound):
            yield actor._get_cert_arn('test')

    @testing.gen_test
    def test_use_cert(self):
        actor = elb_actor.SetCert(
            'Unit Test', {'name': 'unit-test',
                          'region': 'us-east-1',
                          'cert_name': 'unit-cert'}
        )
        elb = mock.Mock()

        yield actor._use_cert(elb=elb, arn='test')
        self.assertEqual(elb.set_listener_SSL_certificate.call_count, 1)

        error = BotoServerError(400, 'test')
        elb.set_listener_SSL_certificate.side_effect = error
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor._use_cert(elb=elb, arn='test')

    @testing.gen_test
    def test_execute(self):
        actor = elb_actor.SetCert(
            'Unit Test', {'name': 'unit-test',
                          'region': 'us-east-1',
                          'cert_name': 'unit-cert'}
        )
        elb = mock.Mock()
        elb.listeners = [
            (443, 443, 'HTTPS', 'HTTPS',
             'arn:aws:iam::12345:server-certificate/nextdoor.com')]
        actor._find_elb = helper.mock_tornado(elb)
        actor._get_cert_arn = helper.mock_tornado('arn')
        actor._check_access = helper.mock_tornado()
        actor._use_cert = helper.mock_tornado()

        yield actor._execute()

        self.assertEqual(actor._check_access._call_count, 0)
        self.assertEqual(actor._use_cert._call_count, 1)

        # Check quick exit if the cert is already in use
        actor._get_cert_arn = helper.mock_tornado(elb.listeners[0][4])
        yield actor._execute()

        # Function calls should remain unchanged
        self.assertEqual(actor._check_access._call_count, 0)
        self.assertEqual(actor._use_cert._call_count, 1)

    @testing.gen_test
    def test_execute_dry(self):
        actor = elb_actor.SetCert(
            'Unit Test', {'name': 'unit-test',
                          'region': 'us-east-1',
                          'cert_name': 'unit-cert'},
            dry=True)

        elb = mock.Mock()
        elb.listeners = [
            (443, 443, 'HTTPS', 'HTTPS',
             'arn:aws:iam::12345:server-certificate/nextdoor.com')]

        actor._find_elb = helper.mock_tornado(elb)
        actor._get_cert_arn = helper.mock_tornado('arn')
        actor._check_access = helper.mock_tornado()

        yield actor._execute()

        self.assertEqual(actor._check_access._call_count, 1)



from boto import exception as boto_exception
from botocore import exceptions as botocore_exceptions
from tornado import concurrent
from tornado import gen
from tornado import queues
from tornado import ioloop

EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class ApiCallQueue:
    """
    Handles queueing up and sending AWS api calls serially,
    with exponential backoff when there is throttling.

    Supports both boto2 and boto3.

    Invoke the `call` method to queue up a new API call.
    """

    def __init__(self):
        self.executor = EXECUTOR

        self._queue = queues.Queue()
        ioloop.IOLoop.current().spawn_callback(self._process_queue)

        # Used for controlling how fast the work queue is processed,
        # with exponential delay on throttling errors.
        self.delay_min = 0.25
        self.delay_max = 30
        # We don't have a delay until we first get throttled.
        self.delay = 0

        # There are a number of different rate limiting messages
        # boto2 can return when rate limits are reached, depending
        # on which apis are used.
        self.boto2_throttle_strings = (
            'Throttling',
            'Rate exceeded',
            'reached max retries',
        )

    @gen.coroutine
    def call(self, api_function, *args, **kwargs):
        """Call a boto2 or boto3 api function.

        Simply invoke this with an api method and its args and kwargs.

        The api function call is coordinated synchronously across all
        calls to this `api_call_queue`, and they will run in order.

        I.e., if you invoke this right after another coroutine invoked this,
        it will block until that other coroutine's call completed.

        If the call ends up being rate limited,
        it will backoff and try again continuously.

        By serializing the api calls to the specific method,
        this prevents a stampeding herd effect that you'd normally get
        with infinite retries.

        There is no limit or timeout on how many times it will retry,
        so in practice this may block an extremely long time if all responses
        are rate limit exceptions.

        Any other failures, like connection timeouts or read timeouts,
        will bubble up immediately and won't be retried here.
        """
        result_queue = queues.Queue(maxsize=1)
        yield self._queue.put((result_queue, api_function, args, kwargs))
        result = yield result_queue.get()
        if isinstance(result, Exception):
            raise result
        raise gen.Return(result)

    @gen.coroutine
    def _process_queue(self):
        """Queue consumer.

        Reads the api functions to call from the internal queue
        along with their individual result queues.
        Calls the api function.
        That result queue is used to pass back the result
        or exception from the call.
        This sleeps between API calls based on `delay`.
        """
        while True:
            result_queue, api_function, args, kwargs = yield self._queue.get()
            try:
                result = yield self._call(api_function, *args, **kwargs)
            except Exception as e:
                result = e
            yield result_queue.put(result)
            yield gen.sleep(self.delay)

    @gen.coroutine
    def _call(self, api_function, *args, **kwargs):
        """Calls the provided api_function in a background thread.

        If the api function returns a response cleanly, this will return it.
        If the api function raises an exception, this raises it up.

        For as long as the api function returns a boto2 or boto3
        rate limiting exception, this will backoff and try again.
        """
        while True:
            try:
                result = yield self._thread(api_function, *args, **kwargs)
                self._decrease_delay()
                raise gen.Return(result)
            except boto_exception.BotoServerError as e:
                # Boto2 exception.
                if e.error_code in self.boto2_throttle_strings:
                    self._increase_delay()
                    yield gen.sleep(self.delay)
                else:
                    self._decrease_delay()
                    raise e
            except botocore_exceptions.ClientError as e:
                # Boto3 exception.
                if e.response['Error']['Code'] == 'Throttling':
                    self._increase_delay()
                    yield gen.sleep(self.delay)
                else:
                    self._decrease_delay()
                    raise e

    def _decrease_delay(self):
        """Decrease `delay` by one step.

        If `delay` is already 0, do nothing.
        If `delay` is `delay_min`, go to 0.

        Otherwise, divide `delay` by 2.
        If that goes below `delay_min`, go to `delay_min`.
        """
        if self.delay == 0:
            return
        if self.delay == self.delay_min:
            self.delay = 0
            return
        self.delay /= 2
        self.delay = max(self.delay, self.delay_min)

    def _increase_delay(self):
        """Increase `delay` by one step.

        If `delay` is already 0, go to `delay_min`.

        Otherwise, multiply `delay` by 2.
        If that goes above `delay_max`, go to `delay_max`.
        """
        if self.delay == 0:
            self.delay = self.delay_min
            return
        self.delay *= 2
        self.delay = min(self.delay, self.delay_max)

    @concurrent.run_on_executor
    def _thread(self, function, *args, **kwargs):
        """Execute `function` in a concurrent thread.

        This allows execution of any function in a thread without having
        to write a wrapper method that is decorated with run_on_executor().
        """
        return function(*args, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.pingdom`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Pingdom actors to pause and unpause checks. These are useful when you are aware
of an expected downtime and don't want to be alerted about it. Also known as
Maintenance mode.

**Required Environment Variables**

:PINGDOM_TOKEN:
  Pingdom API Token

:PINGDOM_USER:
  Pingdom Username (email)

:PINGDOM_PASS:
  Pingdom Password
"""

import logging
import os

from tornado import gen
from tornado import httpclient

from tornado_rest_client import api

from kingpin.constants import REQUIRED
from kingpin.actors import base
from kingpin.actors import exceptions

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


USER = os.getenv('PINGDOM_USER', None)
PASS = os.getenv('PINGDOM_PASS', None)
TOKEN = os.getenv('PINGDOM_TOKEN', None)


class PingdomAPI(api.RestConsumer):

    _ENDPOINT = 'https://api.pingdom.com'
    _CONFIG = {
        'attrs': {
            'checks': {
                'path': '/api/2.0/checks',
                'http_methods': {'get': {}}
            },
            'check': {
                'path': '/api/2.0/checks/%check_id%',
                'http_methods': {'put': {}}
            },
        },
        'auth': {
            'user': USER,
            'pass': PASS
        }
    }


class PingdomClient(api.RestClient):

    # The default exception handling is fine, but the Pingdom API uses a 599 to
    # represent a timeout on the backend of their service.
    _EXCEPTIONS = dict(api.RestClient.EXCEPTIONS)
    _EXCEPTIONS[httpclient.HTTPError]['599'] = None


class PingdomBase(base.BaseActor):

    """Simple Pingdom Abstract Base Object"""

    all_options = {
        'name': (str, REQUIRED, 'Name of the check'),
    }

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(PingdomBase, self).__init__(*args, **kwargs)

        rest_client = PingdomClient(
            headers={'App-Key': TOKEN}
        )
        self._pingdom_client = PingdomAPI(client=rest_client)

    @gen.coroutine
    def _get_check(self):
        """Get check data for actor's option "name".

        Pingdom returns an array of all checks. This method finds the check
        with the exact name and returns its contents.

        Raises InvalidOptions if the check does not exist.
        """
        resp = yield self._pingdom_client.checks().http_get()
        all_checks = resp['checks']
        check = [c for c in all_checks
                 if c['name'] == self.option('name')]

        if not check:
            raise exceptions.InvalidOptions(
                'Check name "%s" was not found.' % self.option('name'))

        raise gen.Return(check[0])


class Pause(PingdomBase):

    """Start Pingdom Maintenance.

    Pause a particular "check" on Pingdom.

    **Options**

    :name:
      (Str) Name of the check

    **Example**

    .. code-block:: json

       { "actor": "pingdom.Pause",
         "desc": "Run Pause",
         "options": {
           "name": "fill-in"
         }
       }

    **Dry run**

    Will assert that the check name exists, but not take any action on it.
    """

    desc = "Pausing check {name}"

    @gen.coroutine
    def _execute(self):
        check = yield self._get_check()

        if self._dry:
            self.log.info('Would pause %s (%s) pingdom check.' % (
                check['name'], check['hostname']))
            raise gen.Return()

        self.log.info('Pausing %s' % check['name'])
        yield self._pingdom_client.check(
            check_id=check['id']).http_put(paused='true')


class Unpause(PingdomBase):

    """Stop Pingdom Maintenance.

    Unpause a particular "check" on Pingdom.

    **Options**

    :name:
      (Str) Name of the check

    **Example**

    .. code-block:: json

       { "actor": "pingdom.Unpause",
         "desc": "Run unpause",
         "options": {
           "name": "fill-in"
         }
       }

    **Dry run**

    Will assert that the check name exists, but not take any action on it.
    """

    desc = "Unpausing check {name}"

    @gen.coroutine
    def _execute(self):
        check = yield self._get_check()

        if self._dry:
            self.log.info('Would unpause %s (%s) pingdom check.' % (
                check['name'], check['hostname']))
            raise gen.Return()

        self.log.info('Unpausing %s' % check['name'])
        yield self._pingdom_client.check(
            check_id=check['id']).http_put(paused='false')



"""Tests for the actors.librato package"""

from nose.plugins.attrib import attr

from tornado import testing

from kingpin.actors import librato
from kingpin.actors import exceptions


__author__ = 'Charles McLaughlin <charles@nextdoor.com>'


class IntegrationLibratoAnnotation(testing.AsyncTestCase):

    """Integration tests against the Librato API.

    These tests actually hit the Librato API and test that the code
    works, as well as validate that the API credentials are working properly.

    Require LIBRATO_TOKEN and LIBRATO_EMAIL environment variables to be set.
    """

    integration = True

    @attr('librato', 'integration', 'integration', 'dry')
    @testing.gen_test
    def integration_test_init_without_token(self):
        # Un-set auth token and make sure the init fails
        librato.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            librato.Annotation(
                'Unit Test Action',
                {'title': 'unittest',
                 'description': 'unittest',
                 'name': 'unittest'}, dry=True)

    @attr('librato', 'integration', 'dry')
    @testing.gen_test
    def integration_test_init_without_email(self):
        # Un-set auth email and make sure the init fails
        librato.EMAIL = None
        with self.assertRaises(exceptions.InvalidCredentials):
            librato.Annotation(
                'Unit Test Action',
                {'title': 'unittest',
                 'description': 'unittest',
                 'name': 'unittest'}, dry=True)

    @attr('librato', 'integration', 'dry')
    @testing.gen_test
    def integration_test_execute_with_invalid_token(self):
        # Set auth token to invalid value and make sure execute fails
        librato.TOKEN = 'Invalid'

        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'}, dry=True)

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()

    @attr('librato', 'integration', 'dry')
    @testing.gen_test
    def integration_test_execute_with_invalid_email(self):
        # Set auth email to invalid value and make sure execute fails
        librato.EMAIL = 'Invalid'

        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'unittest',
             'description': 'unittest',
             'name': 'unittest'}, dry=True)

        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()

    @attr('librato', 'integration', 'dry')
    @testing.gen_test
    def integration_test_execute_dry(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'Kingpin Integration Testing',
             'description': 'Executing integration tests',
             'name': 'kingpin-integration-testing'}, dry=True)

        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('librato', 'integration')
    @testing.gen_test
    def integration_test_execute(self):
        actor = librato.Annotation(
            'Unit Test Action',
            {'title': 'Kingpin Integration Testing',
             'description': 'Executing integration tests',
             'name': 'kingpin-integration-testing'})

        res = yield actor.execute()
        self.assertEqual(res, None)



import copy
import mock
import logging
import json

from tornado import testing
from tornado import httpclient

from kingpin.actors import exceptions
from kingpin.actors import spotinst
from kingpin.actors.test.helper import mock_tornado, tornado_value

__author__ = 'Matt Wise <matt@nextdoor.com>'


class TestSpotinstException(testing.AsyncTestCase):

    def test_no_json_in_body(self):
        fake_body = '400 Bad Mmmkay'
        exc = spotinst.SpotinstException(fake_body)
        self.assertEqual(
            'Unknown error: 400 Bad Mmmkay',
            str(exc))

    def test_invalid_auth_response(self):
        fake_resp_body = mock.MagicMock(name='response_body')
        fake_resp_body.body = json.dumps(
            {
                'request': {
                    'id': 'fake_id',
                    'url': '/fake',
                    'method': 'GET',
                    'timestamp': '2016-12-28T22:36:36.324Z',
                },
                'response': {
                    'error': 'invalid auth',
                    'error_id': 'NOAUTH'
                },
            }
        )
        source_exc = httpclient.HTTPError(
            400, '400 Bad Request', fake_resp_body)
        fake_exc = spotinst.SpotinstException(source_exc)
        self.assertEqual(
            'Spotinst Request ID (fake_id) GET /fake: invalid auth',
            str(fake_exc))

    def test_group_validation_errors(self):
        fake_resp_body = mock.MagicMock(name='response_body')
        fake_resp_body.body = json.dumps(
            {
                'request': {
                    'id': 'fake_id',
                    'url': '/fake',
                    'method': 'GET',
                    'timestamp': '2016-12-28T22:36:36.324Z',
                },
                'response': {
                    'errors': [
                        {'message': 'Cant create spot requests.',
                         'code': 'GENERAL_ERROR'},
                        {'message': 'AMI ami-16fc4976 with an...',
                         'code': 'UnsupportedOperation'},
                    ]
                },
            }
        )
        source_exc = httpclient.HTTPError(
            400, '400 Bad Request', fake_resp_body)
        fake_exc = spotinst.SpotinstException(source_exc)
        self.assertEqual(
            ('Spotinst Request ID (fake_id) GET /fake: GENERAL_ERROR: Cant '
             'create spot requests., UnsupportedOperation: AMI ami-16fc4976 '
             'with an...'),
            str(fake_exc))

    def test_unknown_error_body(self):
        fake_resp_body = mock.MagicMock(name='response_body')
        fake_resp_body.body = json.dumps(
            {
                'request': {
                    'id': 'fake_id',
                    'url': '/fake',
                    'method': 'GET',
                    'timestamp': '2016-12-28T22:36:36.324Z',
                },
                'response': {
                    'something': 'else'
                },
            }
        )
        source_exc = httpclient.HTTPError(
            400, '400 Bad Request', fake_resp_body)
        fake_exc = spotinst.SpotinstException(source_exc)
        self.assertEqual(
            ('Spotinst Request ID (fake_id) GET /fake: '
             '{\'something\': \'else\'}'),
            str(fake_exc))


class TestSpotinstBase(testing.AsyncTestCase):

    """Unit tests for the packagecloud Base actor."""

    def setUp(self, *args, **kwargs):
        super(TestSpotinstBase, self).setUp(*args, **kwargs)
        spotinst.TOKEN = 'Unittest'
        spotinst.DEBUG = True
        spotinst.ACCOUNT_ID = 'act-test'

    def test_init_with_debug_disabled(self):
        spotinst.DEBUG = False
        spotinst.SpotinstBase('Unit Test Action', {})
        self.assertEqual(
            20, logging.getLogger('tornado_rest_client.api').level)

    def test_init_missing_token(self):
        spotinst.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            spotinst.SpotinstBase('Unit Test Action', {})

    def test_init_without_account_id(self):
        spotinst.ACCOUNT_ID = None
        with self.assertRaises(exceptions.InvalidCredentials):
            spotinst.SpotinstBase('Unit Test Action', {})


class TestElastiGroup(testing.AsyncTestCase):

    """Unit tests for the ElastiGroup actor."""

    def setUp(self, *args, **kwargs):
        super(TestElastiGroup, self).setUp(*args, **kwargs)
        file = 'examples/test/spotinst.elastigroup/unittest.json'
        spotinst.TOKEN = 'Unittest'
        spotinst.ACCOUNT_ID = 'act-test'

        # Manually inject some fake values for the subnet/secgrp/zone
        init_tokens = {
            'SECGRP': 'sg-123123',
            'ZONE': 'us-test-1a',
            'SUBNET': 'sn-123123'
        }

        self.actor = spotinst.ElastiGroup(
            'unittest',
            {'name': 'unittest',
             'config': file,
             'wait_on_create': True,
             'wait_on_roll': True},
            init_tokens=init_tokens)
        self.actor._client = mock.Mock()

    def test_init_with_string_roll_settings(self):
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor = spotinst.ElastiGroup(
                options={'name': 'unittest',
                         'config': 'junk',
                         'roll_batch_size': 'some_number'})

    def test_parse_group_config(self):
        self.assertEqual(
            (self.actor._config['group']['compute']
             ['availabilityZones'][0]['name']), 'us-test-1a')
        self.assertEqual(
            self.actor._config['group']['name'], 'unittest')

    def test_parse_group_config_no_config(self):
        self.actor._options['config'] = None
        self.assertEqual(
            None, self.actor._parse_group_config())

    def test_parse_group_config_missing_token(self):
        del(self.actor._init_tokens['ZONE'])
        with self.assertRaises(exceptions.InvalidOptions):
            self.actor._parse_group_config()

    @testing.gen_test
    def test_list_groups(self):
        list_of_groups = {
            'request': {
                'id': 'fake_id',
                'url': '/fake',
                'method': 'GET',
                'timestamp': '2016-12-28T22:36:36.324Z',
            },
            'response': {
                'items': [
                    {'group': {'name': 'test'}}
                ]
            }
        }

        self.actor._client.aws.ec2.list_groups.http_get = mock_tornado(
            list_of_groups)
        ret = yield self.actor._list_groups()
        self.assertEqual(
            ret, [{'group': {'name': 'test'}}])

    @testing.gen_test
    def test_get_group(self):
        matching_group = {
            'name': 'unittest',
            'id': 'bogus',
        }
        self.actor._list_groups = mock_tornado([matching_group])

        ret = yield self.actor._get_group()
        self.assertEqual(ret, {'group': matching_group})

    @testing.gen_test
    def test_get_group_too_many_results(self):
        matching_group = {
            'name': 'unittest',
            'id': 'bogus',
        }
        self.actor._list_groups = mock_tornado(
            [matching_group, matching_group])

        with self.assertRaises(exceptions.InvalidOptions):
            yield self.actor._get_group()

    @testing.gen_test
    def test_get_group_no_groups(self):
        self.actor._list_groups = mock_tornado(None)
        ret = yield self.actor._get_group()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_get_group_no_matching_groups(self):
        unmatching_group = {
            'name': 'unittest-not-matching',
            'id': 'bogus',
        }
        self.actor._list_groups = mock_tornado([unmatching_group])
        ret = yield self.actor._get_group()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_precache(self):
        fake_group = {
            'name': 'unittest',
            'id': 'bogus',
            'group': {
                'capacity': {
                    'target': 128
                }
            }
        }
        self.actor._get_group = mock_tornado(fake_group)
        self.actor._validate_group = mock_tornado(None)
        yield self.actor._precache()

        # First make sure we stored the group
        self.assertEqual(fake_group, self.actor._group)

        # Second, make sure we overwrote the user config's [capacity][target]
        # setting with the spotinst value
        self.assertEqual(self.actor._config['group']['capacity']['target'],
                         128)

    @testing.gen_test
    def test_validate_group(self):
        fake_ret = {'ok': True}
        mock_client = mock.MagicMock()
        mock_client.http_post.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.validate_group = mock_client

        yield self.actor._validate_group()
        mock_client.http_post.assert_called_with(
            group=self.actor._config['group'])

    @testing.gen_test
    def test_get_state(self):
        self.actor._group = True
        ret = yield self.actor._get_state()
        self.assertEqual('present', ret)

    @testing.gen_test
    def test_get_state_false(self):
        self.actor._group = None
        ret = yield self.actor._get_state()
        self.assertEqual('absent', ret)

    @testing.gen_test
    def test_set_state_present(self):
        fake_ret = {
            'ok': True,
            'response': {
                'items': [
                    {'desc': 'new group'}
                ]
            }
        }
        mock_client = mock.MagicMock()
        mock_client.http_post.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.create_group = mock_client

        # Mock out the _get_group method which is called again after the
        # create_group API call is made.
        self.actor._get_group = mock_tornado(None)

        # Mock out the call to our wait_until_stable() call since thats not in
        # the scope of this test.
        self.actor._wait_until_stable = mock_tornado(None)

        yield self.actor._set_state()
        mock_client.http_post.assert_called_with(
            group=self.actor._config['group'])

    @testing.gen_test
    def test_set_state_absent(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group['group']['id'] = 'sig-1234123'
        self.actor._group['group']['createdAt'] = 'timestamp'
        self.actor._group['group']['updatedAt'] = 'timestamp'

        fake_ret = {'ok': True}
        self.actor._options['state'] = 'absent'
        mock_client = mock.MagicMock()
        mock_client.http_delete.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.delete_group.return_value = mock_client

        yield self.actor._set_state()
        mock_client.http_delete.assert_called_with()

    @testing.gen_test
    def test_compare_config(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group['group']['id'] = 'sig-1234123'
        self.actor._group['group']['createdAt'] = 'timestamp'
        self.actor._group['group']['updatedAt'] = 'timestamp'

        # This should return True because the configs are identical..
        ret = yield self.actor._compare_config()
        self.assertEqual(True, ret)

        # Now, lets modify the ElastiGroup config a bit.. the diff should
        # return false.
        self.actor._group['group']['description'] = 'new description'
        ret = yield self.actor._compare_config()
        self.assertEqual(False, ret)

    @testing.gen_test
    def test_compare_config_not_existing(self):
        # Pretend that the group doesn't exist at all in Spotinst
        self.actor._group = None

        # This should return True because the config simply doesnt exist in
        # Spotinst. The _set_state() method will create it during a real run.
        ret = yield self.actor._compare_config()
        self.assertEqual(True, ret)

    @testing.gen_test
    def test_get_config(self):
        self.actor._group = 1
        ret = yield self.actor._get_config()
        self.assertEqual(1, ret)

    @testing.gen_test
    def test_set_config(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group['group']['id'] = 'sig-1234123'
        self.actor._group['group']['createdAt'] = 'timestamp'
        self.actor._group['group']['updatedAt'] = 'timestamp'

        # Pretend to roll the group if a change is made
        self.actor._options['roll_on_change'] = True
        self.actor._roll_group = mock.MagicMock()
        self.actor._roll_group.return_value = tornado_value(None)

        # Mock out the update_group call..
        fake_ret = {
            'response': {
                'items': [
                    {'group': 'object'}
                ]
            }
        }
        mock_client = mock.MagicMock()
        mock_client.http_put.return_value = tornado_value(fake_ret)
        self.actor._client.aws.ec2.update_group.return_value = mock_client

        yield self.actor._set_config()

        mock_client.http_put.assert_called_with(
            group=self.actor._config['group'])
        self.assertEqual(self.actor._group['group'], {'group': 'object'})
        self.actor._roll_group.assert_has_calls([mock.call])

    @testing.gen_test
    def test_roll_group(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group['group']['id'] = 'sig-1234123'
        self.actor._group['group']['createdAt'] = 'timestamp'
        self.actor._group['group']['updatedAt'] = 'timestamp'

        # Mock out the returned calls from Spotinst. The first time we call the
        # roll status method, we'll return no rolls in progress. The second
        # time, there will be a single roll in progress (pretending that the
        # roll_group call was successful), and then the third time it will
        # return no rolls in progress again (pretending that we're done).
        in_progress = {
            "id": "sbgd-44e6d801",
            "status": "in_progress",
            "progress": {
                "unit": "percent",
                "value": 0
            },
            "createdAt": "2017-01-05T15:48:28.000+0000",
            "updatedAt": "2017-01-05T15:49:15.000+0000"
        }
        finished = {
            "id": "sbgd-9f1aa4f6",
            "status": "finished",
            "progress": {
                "unit": "percent",
                "value": 100
            },
            "createdAt": "2017-01-05T15:06:25.000+0000",
            "updatedAt": "2017-01-05T15:22:17.000+0000"
        }

        roll_responses = [
            # First response, no rolls are in progress
            tornado_value({
                "request": {
                    "id": "request-1-no-in-progress",
                    "url": "/aws/ec2/group/sig-b55014f1/roll?limit=5",
                    "method": "GET",
                    "timestamp": "2017-01-05T15:50:44.215Z"
                },
                "response": {
                    "status": {
                        "code": 200,
                        "message": "OK"
                    },
                    "kind": "spotinst:aws:ec2:group:roll",
                    "items": [finished]
                }
            }),

            # Ok now pretend that one roll is in progress, and one is finished
            tornado_value({
                "request": {
                    "id": "request-2-one-in-progress",
                    "url": "/aws/ec2/group/sig-b55014f1/roll?limit=5",
                    "method": "GET",
                    "timestamp": "2017-01-05T15:50:44.215Z"
                },
                "response": {
                    "status": {
                        "code": 200,
                        "message": "OK"
                    },
                    "kind": "spotinst:aws:ec2:group:roll",
                    "items": [in_progress, finished]
                }
            }),

            # Finally, no rolls in progress.
            tornado_value({
                "request": {
                    "id": "request-3-none-in-progress",
                    "url": "/aws/ec2/group/sig-b55014f1/roll?limit=5",
                    "method": "GET",
                    "timestamp": "2017-01-05T15:50:44.215Z"
                },
                "response": {
                    "status": {
                        "code": 200,
                        "message": "OK"
                    },
                    "kind": "spotinst:aws:ec2:group:roll",
                    "items": [finished, finished]
                }
            })
        ]

        # Generate a basic mock object for the entire 'roll' RestConsumer. Mock
        # out the http_get() method to return back the three fake responses
        # above. Mock out the http_put() method to just return safely.
        roll_mock = mock.MagicMock()
        roll_mock.http_get.side_effect = roll_responses
        roll_mock.http_put.side_effect = [tornado_value(None)]
        self.actor._client.aws.ec2.roll.return_value = roll_mock

        # Make the call, and make sure we wait for all roll operations to
        # finish
        self.actor._options['wait_on_roll'] = True
        yield self.actor._roll_group(delay=0.01)

        # Now verify that all the calls were made to the roll_mock
        roll_mock.assert_has_calls([
            mock.call.http_get(),
            mock.call.http_put(gracePeriod=600, batchSizePercentage=20),
            mock.call.http_get(),
            mock.call.http_get(),
        ])

    @testing.gen_test
    def test_wait_until_stable(self):
        # First, lets copy the desired configuration blob. The first test,
        # we'll copy the blob and we'll ensure that they are the same.
        self.actor._group = copy.deepcopy(self.actor._config)

        # Insert some fake data that would normally have been returned in the
        # included blob from Spotinst.
        self.actor._group['group']['id'] = 'sig-1234123'
        self.actor._group['group']['createdAt'] = 'timestamp'
        self.actor._group['group']['updatedAt'] = 'timestamp'

        # Mock out what a pending vs fulfilled instance looks like
        pending = {
            'spotInstanceRequestId': 'sir-n8688grq',
            'instanceId': None,
            'instanceType': 't1.micro',
            'product': 'Linux/UNIX (Amazon VPC)',
            'availabilityZone': 'us-west-2a',
            'createdAt': '2017-01-03T22:30:56.000Z',
            'status': 'pending-evaluation'
        }
        fullfilled = {
            'spotInstanceRequestId': 'sir-n8688grq',
            'instanceId': 'i-abcdefg',
            'instanceType': 't1.micro',
            'product': 'Linux/UNIX (Amazon VPC)',
            'availabilityZone': 'us-west-2a',
            'createdAt': '2017-01-03T22:30:56.000Z',
            'status': 'fullfilled'
        }

        # Create a mock for the group_status API call that returns different
        # results on the 3rd time its called. The first two times, the
        # instances will be in a pending-evaluation state, the third call they
        # will be in a fullfilled state.
        group_status_mock = mock.MagicMock('group_status')
        self.actor._client.aws.ec2.group_status().http_get = group_status_mock
        self.actor._client.aws.ec2.group_status().http_get.side_effect = [
            tornado_value({'response': {'items': [pending, pending]}}),
            tornado_value({'response': {'items': [pending, pending]}}),
            tornado_value({'response': {'items': [fullfilled, fullfilled]}}),
        ]

        # Now make the call
        yield self.actor._wait_until_stable(delay=0.01)
        group_status_mock.assert_has_calls([mock.call(), mock.call(),
                                            mock.call()])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.rightscale.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The RightScale Actors allow you to interact with resources inside your
Rightscale account. These actors all support dry runs properly, but each
actor has its own caveats with ``dry=True``. Please read the instructions
below for using each actor.

**Required Environment Variables**

:RIGHTSCALE_TOKEN:
  RightScale API Refresh Token
  (from the *Account Settings/API Credentials* page)

:RIGHTSCALE_ENDPOINT:
  Your account-specific API Endpoint
  (defaults to https://my.rightscale.com)
"""

from random import randint
import collections
import logging
import os

from tornado import gen
import mock

from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors.utils import dry
from kingpin.actors.rightscale import api

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


TOKEN = os.getenv('RIGHTSCALE_TOKEN', None)
ENDPOINT = os.getenv('RIGHTSCALE_ENDPOINT', 'https://my.rightscale.com')


class ArrayNotFound(exceptions.RecoverableActorFailure):

    """Raised when a ServerArray could not be found."""


class ArrayAlreadyExists(exceptions.RecoverableActorFailure):

    """Raised when a ServerArray already exists by a given name."""


class RightScaleBaseActor(base.BaseActor):

    """Abstract class for creating RightScale cloud actors."""

    account_name = None

    def __init__(self, *args, **kwargs):
        """Initializes the Actor."""
        super(RightScaleBaseActor, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "RIGHTSCALE_TOKEN" environment variable.')

        self._client = api.RightScale(token=TOKEN, endpoint=ENDPOINT)

    @gen.coroutine
    def _find_server_arrays(self, array_name,
                            raise_on='notfound',
                            allow_mock=True,
                            exact=True):
        """Find a ServerArray by name and return it.

        Args:
            array_name: String name of the ServerArray to find.
            raise_on: Either None, 'notfound' or 'found'
            allow_mock: Boolean whether or not to allow a Mock object to be
                        returned instead.
            exact: Boolean whether or not to allow multiple arrays to be
                   returned.

        Raises:
            gen.Return(<rightscale.Resource of Server Array>)
            ArrayNotFound()
            ArrayAlreadyExists()
        """
        if raise_on == 'notfound':
            msg = 'Verifying that array "%s" exists' % array_name
        elif raise_on == 'found':
            msg = 'Verifying that array "%s" does not exist' % array_name
        elif not raise_on:
            msg = 'Searching for array named "%s"' % array_name
        else:
            raise exceptions.UnrecoverableActorFailure(
                'Invalid "raise_on" setting in actor code.')

        self.log.debug(msg)
        array = yield self._client.find_server_arrays(array_name, exact=exact)

        if not array and self._dry and allow_mock:
            # Create a fake ServerArray object thats mocked up to help with
            # execution of the rest of the code.
            self.log.info('Array "%s" not found -- creating a mock.' %
                          array_name)
            array = mock.MagicMock(name=array_name)
            # Give the mock a real identity and give it valid elasticity
            # parameters so the Launch() actor can behave properly.
            array.soul = {
                # Used elsewhere to know whether we're working on a mock
                'fake': True,

                # Fake out common server array object properties
                'name': '<mocked array %s>' % array_name,
                'elasticity_params': {'bounds': {'min_count': 4,
                                                 'max_count': 4}}
            }
            array.self.path = '/fake/array/%s' % randint(10000, 20000)
            array.self.show.return_value = array

        if array and raise_on == 'found':
            raise ArrayAlreadyExists('Array "%s" already exists!' % array_name)

        if not array and raise_on == 'notfound':
            raise ArrayNotFound('Array "%s" not found!' % array_name)

        # Quick note. If many arrays were returned, lets make sure we throw a
        # note to the user so they know whats going on.
        if isinstance(array, list):
            for a in array:
                self.log.info('Matching array found: %s' % a.soul['name'])

        raise gen.Return(array)

    def _generate_rightscale_params(self, prefix, params):
        """Utility function for creating RightScale-style parameters.

        RightScale takes inputs in the form of a hash of key/value pairs, but
        these pairs are in a strange pseudo-dict form. This method takes a
        standard hash and converts it into a rightscale-compatible form.

        For example, take this dict:

            {'name': 'unittest-name',
             'bounds': { 'min_count': 3}

        We return:

            [ ('server_array[name]', 'unittest-name'),
              ('server_array[bounds][min_count]', '3) ]

        For more examples, see our unit tests.

        Args:
            prefix: The key-prefix to use (ie, 'server_array')
            params: The dictionary to squash

        Returns:
            A list of tuples of key/value pairs.
        """
        # Nested loop that compresses a multi level dictinary into a flat
        # array of key=value strings.
        def flatten(d, parent_key=prefix, sep='_'):
            items = []

            if isinstance(d, collections.MutableMapping):
                # If a dict is passed in, break it into its items and
                # then iterate over them.
                for k, v in list(d.items()):
                    new_key = parent_key + '[' + k + ']' if parent_key else k
                    items.extend(flatten(v, new_key))
            elif isinstance(d, list):
                # If an array was passed in, then iterate over the array
                new_key = parent_key + '[]' if parent_key else k
                for item in d:
                    items.extend(flatten(item, new_key))
            else:
                items.append((parent_key, d))

            return items

        return flatten(params)

    @gen.coroutine
    def _get_resource_tags(self, resource):
        tags = yield self._client.get_resource_tags(resource)
        raise gen.Return(tags)

    @gen.coroutine
    @dry('Would have added tags to {resource.soul[name]}')
    def _add_resource_tags(self, resource, tags):
        self.log.info('Adding tags: %s' % ','.join(tags))
        yield self._client.add_resource_tags(resource, tags)

    @gen.coroutine
    @dry('Would have deleted tags from {resource.soul[name]}')
    def _delete_resource_tags(self, resource, tags):
        self.log.info('Removing tags: %s' % ','.join(tags))
        yield self._client.delete_resource_tags(resource, tags)

    @gen.coroutine
    def _ensure_tags(self, res, tags):
        """Ensures that a set of tags are applied to a RightScale resource.

        Args:
            res: The resource object itself
            tags: A list of strings, or a single string as a tag.
        """
        if isinstance(tags, str):
            tags = [tags]

        if not res.href:
            # Must be a mocked-out MCI, meaning its brand new, meaning it has
            # no tags. Definitely set them.
            yield self._add_resource_tags(resource=res, tags=tags)
            self.changed = True
            raise gen.Return()

        existing_tags = yield self._get_resource_tags(res)
        new_tags = tags

        # What tags should we add, delete?
        to_delete = list(set(existing_tags) - set(new_tags))
        to_add = list(set(new_tags) - set(existing_tags))

        if to_delete:
            yield self._delete_resource_tags(resource=res, tags=to_delete)
            self.changed = True
        if to_add:
            yield self._add_resource_tags(resource=res, tags=to_add)
            self.changed = True

    @gen.coroutine
    def _log_account_name(self, *args, **kwargs):
        """Logs out the name of the RightScale account."""
        if not RightScaleBaseActor.account_name:
            ca_resource = self._client._client.cloud_accounts
            cloud_accounts = yield self._client.show(ca_resource)
            account = yield self._client.show(cloud_accounts[0].account)
            RightScaleBaseActor.account_name = account.soul['name']
            log.warning('RightScale account name: %s' % account.soul['name'])

    @gen.coroutine
    def execute(self, *args, **kwargs):
        yield self._log_account_name()
        ret = yield super(RightScaleBaseActor, self).execute(*args, **kwargs)
        raise gen.Return(ret)


class EnsurableRightScaleBaseActor(
        RightScaleBaseActor, base.EnsurableBaseActor):

    """Hacky way to re-use the RightScaleBaseActor but make it ensurable."""

    def __init__(self, *args, **kwargs):
        """Initializes the Actor."""
        super(RightScaleBaseActor, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "RIGHTSCALE_TOKEN" environment variable.')

        self._client = api.RightScale(token=TOKEN, endpoint=ENDPOINT)
        self._gather_methods()



"""Simple integration tests for the AWS SQS actors."""

from nose.plugins.attrib import attr
import logging
import uuid

from tornado import testing

from kingpin import utils
from kingpin.actors.aws import settings
from kingpin.actors.aws import sqs
import importlib

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

# Generate a common UUID for this particular set of tests
UUID = uuid.uuid4().hex

log = logging.getLogger(__name__)
logging.getLogger('boto').setLevel(logging.INFO)


class IntegrationSQS(testing.AsyncTestCase):

    """High level SQS Actor testing.

    This suite of tests performs the following actions:
    * Create a queue with a randomized name
    * Add a few messages to the queue
    * Start the WaitUntilEmpty task
    * Check that this task is not exiting while there are messages
    * Remove all the messages from the queue
    * Check that the WaitUntilEmpty notices, and returns success
    * Delete the temporary queue

    Note, these tests must be run in-order. The order is defined by
    their definition order in this file. Nose follows this order according
    to its documentation:

        http://nose.readthedocs.org/en/latest/writing_tests.html
    """

    integration = True

    queue_name = 'integration-test-%s' % UUID
    region = 'us-east-1'

    @attr('aws', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_01a_create_queue_dry(self):
        actor = sqs.Create(
            'Create %s' % self.queue_name,
            {'name': self.queue_name,
             'region': self.region},
            dry=True)
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_01b_create_queue(self):
        actor = sqs.Create(
            'Create %s' % self.queue_name,
            {'name': self.queue_name,
             'region': self.region})
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_02a_monitor_queue_dry(self):
        actor = sqs.WaitUntilEmpty('Wait until empty',
                                   {'name': self.queue_name,
                                    'region': self.region},
                                   dry=True)
        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=60)
    def integration_02b_monitor_queue(self):

        actor = sqs.WaitUntilEmpty('Wait until empty',
                                   {'name': self.queue_name,
                                    'region': self.region})

        log.debug('New queue should be empty')
        queue = actor.sqs_conn.get_queue(self.queue_name)
        self.assertEqual(queue.count(), 0)

        done = yield actor.execute()
        yield utils.tornado_sleep()
        self.assertEqual(done, None)

    @attr('aws', 'integration', 'dry')
    @testing.gen_test()
    def integration_03a_delete_queue_dry(self):
        actor = sqs.Delete('Delete %s' % self.queue_name,
                           {'name': self.queue_name,
                            'region': self.region,
                            'idempotent': True},
                           dry=True)

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)  # Delete actor sleeps and retries.
    def integration_03b_delete_queue(self):
        actor = sqs.Delete('Delete %s' % self.queue_name,
                           {'name': self.queue_name,
                            'region': self.region})

        done = yield actor.execute()
        self.assertEqual(done, None)

    @attr('aws', 'integration')
    @testing.gen_test(timeout=120)  # Delete actor sleeps and retries.
    def integration_03c_delete_fake_queue(self):
        settings.SQS_RETRY_DELAY = 0
        importlib.reload(sqs)
        actor = sqs.Delete('Delete %s' % self.queue_name,
                           {'name': 'totally-fake-queue',
                            'region': self.region})

        with self.assertRaises(sqs.QueueNotFound):
            yield actor.execute()
